from mdps_ds_lib.stage_in_out.download_granules_abstract import DownloadGranulesAbstract
import logging
import os

from mdps_ds_lib.lib.aws.aws_s3 import AwsS3

LOGGER = logging.getLogger(__name__)


class DownloadGranulesS3(DownloadGranulesAbstract):

    def __init__(self) -> None:
        super().__init__()
        self.__s3 = AwsS3()

    def _set_props_from_env(self):
        missing_keys = [k for k in [self.STAC_JSON, self.DOWNLOAD_DIR_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        self._retrieve_stac_json()
        self._setup_download_dir()
        return self

    def _download_one_item(self, downloading_url):
        local_file_path = self.__s3.set_s3_url(downloading_url).download(self._download_dir)
        return local_file_path



import logging
from typing import Union

from botocore.exceptions import ClientError

from mdps_ds_lib.lib.aws.aws_cred import AwsCred

LOGGER = logging.getLogger(__name__)


class AwsParamStore(AwsCred):
    def __init__(self):
        super().__init__()
        self.__ssm_client = self.get_client('ssm')
        self.__ssm = None

    def set_param(self, key: str, val: str, encrypted: bool = True):
        self.__ssm_client.put_parameter(Name=key, Value=val,
                                        Type='SecureString' if encrypted else 'String',
                                        Overwrite=True)

    def get_param(self, param_name: str) -> Union[str, None]:
        """
        Ref: https://stackoverflow.com/questions/46063138/how-can-i-import-the-boto3-ssm-parameternotfound-exception
            on how to catch ParameterNotFound exception

        :param param_name: named of parameter in Parameter store
        :return: plain value of the parameter or None if some exception.
        """
        try:
            param_response = self.__ssm_client.get_parameter(Name=param_name, WithDecryption=True)
            if 'ResponseMetadata' not in param_response or \
                    'HTTPStatusCode' not in param_response['ResponseMetadata'] or \
                    param_response['ResponseMetadata']['HTTPStatusCode'] != 200 or \
                    'Parameter' not in param_response or \
                    'Value' not in param_response['Parameter']:
                return None
        except ClientError:
            LOGGER.exception('cannot get parameter store value for %s', param_name)
            return None
        return param_response['Parameter']['Value']



import logging
import calendar
import time
from datetime import datetime, timezone, timedelta
from time import mktime
import re

module_log = logging.getLogger(__name__)


class TimeUtils:
    REG_IS_NUMERIC = re.compile("^\d+?\.\d+?$")
    DOY = re.compile("^\d{4}-\d{3}T\d{2}:\d{2}:\d{2}(\.\d{1,9})?$")
    MMDD = re.compile("^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,9})?$")

    DOY_FORMAT = '%Y-%jT%H:%M:%S'
    MMDD_FORMAT = '%Y-%m-%dT%H:%M:%S'
    GB_1 = 1000000000
    YR_IN_SECOND = 31536000

    def __init__(self):
        self.__time_obj = datetime.utcnow()

    def parse_from_str(self, timestamp_str: str, fmt='%Y-%m-%dT%H:%M:%S%z', in_ms=False):
        self.__time_obj = datetime.strptime(timestamp_str, fmt)
        return self

    def parse_from_unix(self, unix_timestamp, in_ms=False):
        converting_timestamp = unix_timestamp / 1000 if in_ms is True else unix_timestamp
        self.__time_obj = datetime.fromtimestamp(converting_timestamp, timezone(timedelta(0, 0, 0, 0)))
        return self

    def set_datetime_obj(self, time_obj: datetime):
        self.__time_obj = time_obj
        return self

    def get_datetime_obj(self):
        return self.__time_obj

    def get_datetime_unix(self, in_ms=False):
        return int(self.__time_obj.timestamp()) if not in_ms else int(self.__time_obj.timestamp() * 1000)

    def get_datetime_str(self, fmt='%Y-%m-%dT%H:%M:%S %z', in_ms=True):
        dt_str = self.__time_obj.strftime(fmt)
        dt_str = dt_str.replace('0000', '00:00') if fmt.endswith(' %z') else dt_str
        return dt_str

    @staticmethod
    def get_current_year():
        return datetime.utcnow().year

    @staticmethod
    def is_smaller(target, source):
        """
        if left is smaller than right

        :param target: str or int - the target value
        :param source: str or int - the source value
        :return: bool - if left is smaller than right
        """
        if isinstance(target, str):
            target = TimeUtils.encode_datetime(target)
        if isinstance(source, str):
            source = TimeUtils.encode_datetime(source)
        return target < source

    @staticmethod
    def is_between(target, smaller, larger):
        """
        if target is between smaller and larger

        NOTE: smaller and larger are not validated

        :param target: str or int - the value to be compared
        :param smaller: str or int - the smaller value
        :param larger: str or int - the larger value
        :return: bool - if target is between small and larger value
        """
        if isinstance(target, str):
            target = TimeUtils.encode_datetime(target)
        if isinstance(smaller, str):
            smaller = TimeUtils.encode_datetime(smaller)
        if isinstance(larger, str):
            larger = TimeUtils.encode_datetime(larger)
        return smaller <= target <= larger

    @staticmethod
    def get_current_unix_milli():
        return int(datetime.now().timestamp() * 10**3)

    @staticmethod
    def get_current_unix_nano():
        """
        Get the current unix time in nanosecond.

        Note-1: datetime.utcnow() will get the current utc.
                but datetime.utcnow().timestamp() will assume take utc time as input,
                it will assumes it is in current timezone, and convert to utc.
                Hence, only datetime.now() is needed.
        Note-2: need to convert to int. if not, it returns in float form.
        :return:
        """
        return int(datetime.now().timestamp() * 10**9)

    @staticmethod
    def get_unix_from_timestamp(str_timestamp):
        if TimeUtils.REG_IS_NUMERIC.match(str_timestamp):
            return str_timestamp
        try:
            dt = datetime.strptime(str_timestamp, '%Y-%jT%H:%M:%S.%f')  # yyyy-DDD'T'HH:mm:ss.SSS
        except Exception as e:
            module_log.error('Date conversion failed for %s' % str_timestamp)
            raise e

        assert not dt.tzinfo, "Datetime stamps must use UTC time."
        return mktime(dt.timetuple())

    @staticmethod
    def reduce_to_ms(unix_time):
        return divmod(unix_time, 10 ** 6)[0]

    @staticmethod
    def decode_datetime(unix_time, doy=True):
        """
        convert unix time (milliseconds) to YYYY-DDDTHH:mm:ss.SSS in UTC
        it assumes that the unix time is in milliseconds.
        :param unix_time: unix time in millisecond
        :param doy: decoding in DOY format? defaulted it to True
        :return: YYYY-DDDTHH:mm:ss.SSS or YYYY-MM-DDTHH:mm:SS
        """
        str_format = TimeUtils.DOY_FORMAT if doy is True else TimeUtils.MMDD_FORMAT
        return '{0}.{1:09d}'.format(datetime.utcfromtimestamp(unix_time // TimeUtils.GB_1).strftime(str_format),
                                    unix_time % TimeUtils.GB_1)
        pass

    @staticmethod
    def __append_zeros(nanosecond):
        if len(nanosecond) == 9:  # already in nanosecond format
            return nanosecond

        if len(nanosecond) > 9:  # greater than nanosecond?
            return nanosecond[0:9]  # for now, we only accept nanosecond

        return nanosecond + ''.zfill(9 - len(nanosecond))

        pass

    @staticmethod
    def encode_datetime(to_be_parsed):
        """
        convert datetime string to unix time.
        Assumption: the unix time will be in NANOseconds
        accepted formats
        'YYYY-DDDTHH:mm:ss.SZ' S can vary from 1 - 9
        'YYYY-DDDTHH:mm:ss.SSS'
        'YYYY-DDDTHH:mm:ss.SSSSSSZ'
        'YYYY-DDDTHH:mm:ss.SSSSSS'
        'YYYY-DDDTHH:mm:ss'
        calendar.timegm will convert the string datetime to UTC time.
        But it will not include millisecond.
        Temporary workaround is to manually add them.
        :param to_be_parsed: datetime in string to be parsed
        :return: unix epoch including millisecond
        """
        if TimeUtils.REG_IS_NUMERIC.match(to_be_parsed):  # if it is double, return as it is
            return to_be_parsed

        if to_be_parsed.endswith('Z') or to_be_parsed.endswith('z'):  # ending with Z, remove it
            to_be_parsed = to_be_parsed[:-1]

        if TimeUtils.DOY.match(to_be_parsed):  # Day-Of-year format
            matching_format = '%Y-%jT%H:%M:%S.%f'
        elif TimeUtils.MMDD.match(to_be_parsed):  # MM-dd format
            matching_format = '%Y-%m-%dT%H:%M:%S.%f'
        else:  # not supported yet
            module_log.error('Date conversion failed for %s' % to_be_parsed)
            return None

        nanosecond = 0
        if '.' in to_be_parsed:  # it has floating points
            split_milli = to_be_parsed.split('.')  # split to datetime & floating points
            if len(split_milli) == 2:  # if the result array has 2,
                nanosecond += int(TimeUtils.__append_zeros(split_milli[1]))
            # if there are more than 6 decimal values, the `time` library cannot handle it.
            # replacing the fraction with 0
            to_be_parsed = split_milli[0] + '.0'
        else:  # it has no floating points
            to_be_parsed += '.0'  # adding 0 nanosecond for regex

        try:
            return calendar.timegm(time.strptime(to_be_parsed, matching_format)) * TimeUtils.GB_1 \
                   + nanosecond
        except Exception as e:
            module_log.error('Date conversion failed for {}. exception: {}'.format(to_be_parsed, e))
            return None

    @staticmethod
    def get_current_time(doy_format=False, include_time=True, include_fraction=True):
        output_format = '%Y-%j' if doy_format is True else '%Y-%m-%d'
        if include_time:
            output_format = '{}T%H:%M:%S'.format(output_format)
            if include_fraction:
                output_format = '{}.%f'.format(output_format)
        return datetime.utcnow().strftime(output_format)
    pass


import json
import logging
import os
from abc import ABC, abstractmethod
from multiprocessing import Manager

import requests
from mdps_ds_lib.lib.cognito_login.cognito_token_retriever import CognitoTokenRetriever

from mdps_ds_lib.lib.constants import Constants
from pystac import ItemCollection, Asset, Item

from mdps_ds_lib.lib.processing_jobs.job_executor_abstract import JobExecutorAbstract
from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerProps
from mdps_ds_lib.lib.processing_jobs.job_manager_memory import JobManagerMemory
from mdps_ds_lib.lib.processing_jobs.multithread_processor import MultiThreadProcessorProps, \
    MultiThreadProcessor
from mdps_ds_lib.lib.utils.file_utils import FileUtils

LOGGER = logging.getLogger(__name__)


class DownloadItemExecutor(JobExecutorAbstract):
    def __init__(self, downloading_keys, downloading_roles, download_one_item_func, result_list, error_list) -> None:
        self._downloading_keys = downloading_keys
        self._downloading_roles = downloading_roles
        self._download_one_item_func = download_one_item_func
        self.__result_list = result_list
        self.__error_list = error_list

    def validate_job(self, job_obj):
        return isinstance(job_obj, Item)

    def __is_downloading_this_file(self, file_key, file_role):
        if len(self._downloading_roles) > 0:
            role_comparison_result = file_role in self._downloading_roles
            LOGGER.debug(f'Result for role comparison for {file_key}, {file_role} v. {self._downloading_roles}. Result: {role_comparison_result}')
            return role_comparison_result
        if len(self._downloading_keys) > 0:
            key_comparison_result = file_key in self._downloading_keys
            LOGGER.debug(f'Result for role comparison for {file_key}, {file_role} v. {self._downloading_keys}. Result: {key_comparison_result}')
            return key_comparison_result
        LOGGER.debug(f'both _downloading_keys and _downloading_roles are not set. Downloading: {file_key}, {file_role}')
        return True

    def execute_job(self, granule_item, lock) -> bool:
        try:
            new_asset_dict = {}
            for name, value_dict in granule_item.assets.items():
                value_dict: Asset = value_dict
                role = value_dict.roles[0] if value_dict.roles is not None and len(value_dict.roles) > 0 else 'unknown'
                if not self.__is_downloading_this_file(name, role):
                    LOGGER.debug(f'skipping {name}. Not in downloading keys')
                    continue
                downloading_url = value_dict.href
                LOGGER.debug(f'downloading: {downloading_url}')
                self._download_one_item_func(downloading_url)
                value_dict.href = os.path.join('', os.path.basename(downloading_url))
                new_asset_dict[name] = value_dict
            granule_item.assets = new_asset_dict
            self.__result_list.put(granule_item)
        except Exception as e:
            LOGGER.exception(f'error downloading granule: {granule_item.id}')
            self.__error_list.put({'error': str(e), 'id': granule_item.id, })
        LOGGER.debug(f'done DownloadItemExecutor#execute_job')
        return True  # always return true?


class DownloadGranulesAbstract(ABC):
    STAC_AUTH_TYPE = 'STAC_AUTH_TYPE'
    STAC_JSON = 'STAC_JSON'
    DOWNLOAD_DIR_KEY = 'DOWNLOAD_DIR'
    DOWNLOADING_KEYS = 'DOWNLOADING_KEYS'
    DOWNLOADING_ROLES = 'DOWNLOADING_ROLES'
    VERIFY_SSL_KEY = 'VERIFY_SSL_KEY'

    def __init__(self) -> None:
        super().__init__()
        self._granules_json: ItemCollection = {}
        self._download_dir = '/tmp'
        self._downloading_roles = os.environ.get(self.DOWNLOADING_ROLES, '')
        self._downloading_roles = set([]) if self._downloading_roles == '' else set([k.strip() for k in self._downloading_roles.strip().split(',')])
        self._downloading_keys = os.environ.get(self.DOWNLOADING_KEYS, '')
        self._downloading_keys = set([]) if self._downloading_keys == '' else set([k.strip() for k in self._downloading_keys.strip().split(',')])
        self._parallel_count = int(os.environ.get(Constants.PARALLEL_COUNT, '-1'))

    @abstractmethod
    def _set_props_from_env(self):
        raise NotImplementedError(f'to be implemented in concrete classes')

    @abstractmethod
    def _download_one_item(self, downloading_url):
        raise NotImplementedError(f'to be implemented in concrete classes')

    def _setup_download_dir(self):
        self._download_dir = os.environ.get(self.DOWNLOAD_DIR_KEY)
        self._download_dir = self._download_dir[:-1] if self._download_dir.endswith('/') else self._download_dir
        LOGGER.debug(f'creating download dir: {self._download_dir}')
        FileUtils.mk_dir_p(self._download_dir)
        return self

    def _retrieve_stac_json(self):
        raw_stac_json = os.environ.get(self.STAC_JSON)
        LOGGER.debug(f'attempting to decode raw_stac_json to JSON object')
        try:
            self._granules_json = ItemCollection.from_dict(json.loads(raw_stac_json))
            return self
        except:
            LOGGER.debug(f'raw_stac_json is not STAC_JSON: {raw_stac_json}. trying to see if file exists')

        if raw_stac_json.startswith('https'):
            if os.environ.get(self.STAC_AUTH_TYPE, 'NONE').strip().upper() == 'UNITY':
                LOGGER.debug(f'STAC_AUTH_TYPE = UNITY')
                token_retriever = CognitoTokenRetriever()
                token = token_retriever.start()
                header = {'Authorization': f'Bearer {token}'}
                verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
                downloading_response = requests.get(url=raw_stac_json, headers=header, verify=verify_ssl)
            else:
                LOGGER.debug(f'download raw stac json from public URL: {raw_stac_json}')
                downloading_response = requests.get(raw_stac_json)
            downloading_response.raise_for_status()
            self._granules_json = ItemCollection.from_dict(json.loads(downloading_response.content.decode()))
            return self
        if not FileUtils.file_exist(raw_stac_json):
            raise ValueError(f'missing file or not JSON: {raw_stac_json}')
        json_stac = FileUtils.read_json(raw_stac_json)
        if json_stac is None:
            raise ValueError(f'{raw_stac_json} is not JSON')
        self._granules_json = ItemCollection.from_dict(json_stac)
        return self
    
    def download(self, **kwargs) -> str:
        self._set_props_from_env()
        LOGGER.debug(f'creating download dir: {self._download_dir}')
        if len(self._granules_json.items) < 1:
            LOGGER.warning(f'cannot find any granules')
            granules_json_dict = self._granules_json.to_dict(False)
            FileUtils.write_json(os.path.join(self._download_dir, 'downloaded_feature_collection.json'), granules_json_dict, overwrite=True, prettify=True)
            return json.dumps(granules_json_dict)
        # local_items = []
        # error_list = []
        local_items = Manager().Queue()
        error_list = Manager().Queue()
        job_manager_props = JobManagerProps()
        for each_item in self._granules_json.items:
            job_manager_props.memory_job_dict[each_item.id] = each_item

        # https://www.infoworld.com/article/3542595/6-python-libraries-for-parallel-processing.html
        multithread_processor_props = MultiThreadProcessorProps(self._parallel_count)
        multithread_processor_props.job_manager = JobManagerMemory(job_manager_props)
        multithread_processor_props.job_executor = DownloadItemExecutor(self._downloading_keys,self._downloading_roles, self._download_one_item, local_items, error_list)
        multithread_processor = MultiThreadProcessor(multithread_processor_props)
        multithread_processor.start()

        LOGGER.debug(f'finished downloading all granules')
        local_items_list = []
        while not local_items.empty():
            local_items_list.append(local_items.get())

        error_list_list = []
        while not error_list.empty():
            error_list_list.append(error_list.get())

        self._granules_json.items = local_items_list
        LOGGER.debug(f'writing features collection json to downloading directory')
        granules_json_dict = self._granules_json.to_dict(transform_hrefs=False)
        FileUtils.write_json(os.path.join(self._download_dir, 'downloaded_feature_collection.json'), granules_json_dict, overwrite=True, prettify=True)
        LOGGER.debug(f'writing errors if any')
        if len(error_list_list) > 0:
            with open(f'{self._download_dir}/error.log', 'w') as error_file:
                error_file.write(json.dumps(error_list_list, indent=4))
        return json.dumps(granules_json_dict)



import json
from urllib.parse import unquote

from mdps_ds_lib.lib.utils.json_validator import JsonValidator


class AwsMessageTransformers:
    SQS_SCHEMA = {
        'type': 'object',
        'properties': {
            'Records': {
                'type': 'array',
                'minItems': 1,
                'maxItems': 1,
                'items': {
                    'type': 'object',
                    'properties': {
                        'body': {'type': 'string', 'minLength': 1}
                    },
                    'required': ['body']
                }
            }
        },
        'required': ['Records']
    }
    SNS_EVENT_SCHEMA = {
        'type': 'object',
        'properties': {
            'Records': {
                'type': 'array',
                'minItems': 1,
                'maxItems': 1,
                'items': {
                    'type': 'object',
                    'properties': {
                        'Sns': {'type': 'object'}
                    },
                    'required': ['Sns']
                }
            }
        },
        'required': ['Records']
    }

    SNS_SCHEMA = {
        "type": "object",
        "properties": {
            "Type": {"type": "string"},
            "MessageId": {"type": "string"},
            "TopicArn": {"type": "string"},
            # "Subject": {"type": "string"},
            "Timestamp": {"type": "string"},
            "SignatureVersion": {"type": "string"},
            "Signature": {"type": "string"},
            "SigningCertURL": {"type": "string"},
            "UnsubscribeURL": {"type": "string"},
            "Message": {"type": "string"},
        },
        "required": ["Message"]
    }

    S3_RECORD_SCHEMA = {
        'type': 'object',
        'properties': {'Records': {
            'type': 'array',
            'minItems': 1,
            'maxItems': 1,
            'items': {
                'type': 'object',
                'properties': {
                    'eventName': {'type': 'string'},
                    's3': {
                        'type': 'object',
                        'properties': {
                            'bucket': {
                                'type': 'object',
                                'properties': {'name': {'type': 'string', 'minLength': 1}},
                                'required': ['name']
                            },
                            'object': {
                                'type': 'object',
                                'properties': {'key': {'type': 'string', 'minLength': 1}},
                                'required': ['key']
                            }},
                        'required': ['bucket', 'object']
                    }
                },
                'required': ['eventName', 's3']
            }
        }},
        'required': ['Records']
    }

    def sqs_sns(self, raw_msg: json):
        result = JsonValidator(self.SQS_SCHEMA).validate(raw_msg)
        if result is not None:
            raise ValueError(f'input json has SQS validation errors: {result}')
        sqs_msg_body = raw_msg['Records'][0]['body']
        sqs_msg_body = json.loads(sqs_msg_body)
        result = JsonValidator(self.SNS_SCHEMA).validate(sqs_msg_body)
        if result is not None:
            raise ValueError(f'input json has SNS validation errors: {result}')
        sns_msg_body = sqs_msg_body['Message']
        sns_msg_body = json.loads(sns_msg_body)
        return sns_msg_body

    def get_message_from_sns_event(self, raw_msg: json):
        result = JsonValidator(self.SNS_EVENT_SCHEMA).validate(raw_msg)
        if result is not None:
            raise ValueError(f'input json has SNS_EVENT_SCHEMA validation errors: {result}')
        sns_msg = raw_msg['Records'][0]['Sns']
        result = JsonValidator(self.SNS_SCHEMA).validate(sns_msg)
        if result is not None:
            raise ValueError(f'input json has SNS validation errors: {result}')
        sns_msg_body = sns_msg['Message']
        sns_msg_body = json.loads(sns_msg_body)
        return sns_msg_body

    def get_s3_from_sns(self, sns_msg_body):
        result = JsonValidator(self.S3_RECORD_SCHEMA).validate(sns_msg_body)
        if result is not None:
            raise ValueError(f'sns_msg_body did not pass S3_RECORD_SCHEMA: {result}')
        s3_summary = {
            'eventName': sns_msg_body['Records'][0]['eventName'],
            'bucket': sns_msg_body['Records'][0]['s3']['bucket']['name'],
            'key': unquote(sns_msg_body['Records'][0]['s3']['object']['key'].replace('+', ' ')),
        }
        return s3_summary


import json
import os

from pystac import Item, Asset, Link
from pystac.utils import datetime_to_str

from mdps_ds_lib.lib.cumulus_stac.stac_transformer_abstract import StacTransformerAbstract
from mdps_ds_lib.lib.utils.json_validator import JsonValidator
from mdps_ds_lib.lib.utils.time_utils import TimeUtils

STAC_ITEM_SCHEMA = '''{
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "https://schemas.stacspec.org/v1.0.0/item-spec/json-schema/item.json#",
    "title": "STAC Item",
    "type": "object",
    "description": "This object represents the metadata for an item in a SpatioTemporal Asset Catalog.",
    "allOf": [
        {
            "$ref": "#/definitions/core"
        }
    ],
    "definitions": {
        "common_metadata": {
            "allOf": [
                {
                    "$ref": "basics.json"
                },
                {
                    "$ref": "datetime.json"
                },
                {
                    "$ref": "instrument.json"
                },
                {
                    "$ref": "licensing.json"
                },
                {
                    "$ref": "provider.json"
                }
            ]
        },
        "core": {
            "allOf": [
                {
                    "$ref": "https://geojson.org/schema/Feature.json"
                },
                {
                    "oneOf": [
                        {
                            "type": "object",
                            "required": [
                                "geometry",
                                "bbox"
                            ],
                            "properties": {
                                "geometry": {
                                    "$ref": "https://geojson.org/schema/Geometry.json"
                                },
                                "bbox": {
                                    "type": "array",
                                    "oneOf": [
                                        {
                                            "minItems": 4,
                                            "maxItems": 4
                                        },
                                        {
                                            "minItems": 6,
                                            "maxItems": 6
                                        }
                                    ],
                                    "items": {
                                        "type": "number"
                                    }
                                }
                            }
                        },
                        {
                            "type": "object",
                            "required": [
                                "geometry"
                            ],
                            "properties": {
                                "geometry": {
                                    "type": "null"
                                },
                                "bbox": {
                                    "not": {}
                                }
                            }
                        }
                    ]
                },
                {
                    "type": "object",
                    "required": [
                        "stac_version",
                        "id",
                        "links",
                        "assets",
                        "properties"
                    ],
                    "properties": {
                        "stac_version": {
                            "title": "STAC version",
                            "type": "string",
                            "const": "1.0.0"
                        },
                        "stac_extensions": {
                            "title": "STAC extensions",
                            "type": "array",
                            "uniqueItems": true,
                            "items": {
                                "title": "Reference to a JSON Schema",
                                "type": "string",
                                "format": "iri"
                            }
                        },
                        "id": {
                            "title": "Provider ID",
                            "description": "Provider item ID",
                            "type": "string",
                            "minLength": 1
                        },
                        "links": {
                            "title": "Item links",
                            "description": "Links to item relations",
                            "type": "array",
                            "items": {
                                "$ref": "#/definitions/link"
                            }
                        },
                        "assets": {
                            "$ref": "#/definitions/assets"
                        },
                        "properties": {
                            "allOf": [
                                {
                                    "$ref": "#/definitions/common_metadata"
                                },
                                {
                                    "anyOf": [
                                        {
                                            "required": [
                                                "datetime"
                                            ],
                                            "properties": {
                                                "datetime": {
                                                    "not": {
                                                        "type": "null"
                                                    }
                                                }
                                            }
                                        },
                                        {
                                            "required": [
                                                "datetime",
                                                "start_datetime",
                                                "end_datetime"
                                            ]
                                        }
                                    ]
                                }
                            ]
                        }
                    },
                    "if": {
                        "properties": {
                            "links": {
                                "contains": {
                                    "required": [
                                        "rel"
                                    ],
                                    "properties": {
                                        "rel": {
                                            "const": "collection"
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "then": {
                        "required": [
                            "collection"
                        ],
                        "properties": {
                            "collection": {
                                "title": "Collection ID",
                                "description": "The ID of the STAC Collection this Item references to.",
                                "type": "string",
                                "minLength": 1
                            }
                        }
                    },
                    "else": {
                        "properties": {
                            "collection": {
                                "not": {}
                            }
                        }
                    }
                }
            ]
        },
        "link": {
            "type": "object",
            "required": [
                "rel",
                "href"
            ],
            "properties": {
                "href": {
                    "title": "Link reference",
                    "type": "string",
                    "format": "iri-reference",
                    "minLength": 1
                },
                "rel": {
                    "title": "Link relation type",
                    "type": "string",
                    "minLength": 1
                },
                "type": {
                    "title": "Link type",
                    "type": "string"
                },
                "title": {
                    "title": "Link title",
                    "type": "string"
                }
            }
        },
        "assets": {
            "title": "Asset links",
            "description": "Links to assets",
            "type": "object",
            "additionalProperties": {
                "$ref": "#/definitions/asset"
            }
        },
        "asset": {
            "allOf": [
                {
                    "type": "object",
                    "required": [
                        "href"
                    ],
                    "properties": {
                        "href": {
                            "title": "Asset reference",
                            "type": "string",
                            "format": "iri-reference",
                            "minLength": 1
                        },
                        "title": {
                            "title": "Asset title",
                            "type": "string"
                        },
                        "description": {
                            "title": "Asset description",
                            "type": "string"
                        },
                        "type": {
                            "title": "Asset type",
                            "type": "string"
                        },
                        "roles": {
                            "title": "Asset roles",
                            "type": "array",
                            "items": {
                                "type": "string"
                            }
                        }
                    }
                },
                {
                    "$ref": "#/definitions/common_metadata"
                }
            ]
        }
    }
}'''


CUMULUS_FILE_SCHEMA = {
    "type": "object",
    "required": ["bucket", "key", "fileName", "type", ],
    "properties": {
        "bucket": {"type": "string"},
        "key": {"type": "string"},
        "fileName": {"type": "string"},
        "type": {"type": "string"},
    }
}


class ItemTransformer(StacTransformerAbstract):
    def __init__(self):
        super().__init__()
        self.__stac_item_schema = json.loads(STAC_ITEM_SCHEMA)
        self.__cumulus_granule_schema = {}
        self.CUMULUS_2_STAC_KEYS_MAP = {
            'beginningDateTime': 'start_datetime',
            'endingDateTime': 'end_datetime',
            'productionDateTime': 'created',
            'updatedAt': 'updated',
            'granuleId': 'id',
            'collectionId': 'collection',
        }
        self.STAC_2_CUMULUS_KEYS_MAP = {v: k for k, v in self.CUMULUS_2_STAC_KEYS_MAP.items()}

    def __get_asset_name(self, input_dict):
        return os.path.basename(input_dict['fileName'])
        # if input_dict['type'] == 'data':
        #     return 'data'
        # if input_dict['type'] == 'metadata':
        #     filename = input_dict['fileName'].upper().strip()
        #     if filename.endswith('.CMR.XML'):
        #         return 'metadata__cmr'
        #     if filename.endswith('.PDS.XML'):
        #         return 'metadata__xml'
        #     return 'metadata__data'
        # return input_dict['type']

    def __get_asset_obj(self, input_dict):
        """
      {
        "bucket": "uds-sbx-cumulus-staging",
        "checksum": "9817be382b87c48ebe482b9c47d1525a",
        "checksumType": "md5",
        "fileName": "test_file01.cmr.xml",
        "key": "URN:NASA:UNITY:UDS_LOCAL_TEST:DEV:UDS_COLLECTION___2311091417/URN:NASA:UNITY:UDS_LOCAL_TEST:DEV:UDS_COLLECTION___2311091417:test_file01/test_file01.cmr.xml",
        "size": 1768,
        "source": "s3://uds-staging/URN:NASA:UNITY:UDS_LOCAL_TEST:DEV:UDS_COLLECTION___2311091417/URN:NASA:UNITY:UDS_LOCAL_TEST:DEV:UDS_COLLECTION___2311091417:test_file01/test_file01.cmr.xml",
        "type": "metadata"
      }
        "bucket": "am-uds-dev-cumulus-internal",
        "key": "ATMS_SCIENCE_Group___1/P1570515ATMSSCIENCEAAT16032024518500.PDS",
        "size": 760,
        "fileName": "P1570515ATMSSCIENCEAAT16032024518500.PDS",
        "source": "data/SNPP_ATMS_Level0_T/ATMS_SCIENCE_Group/2016/031//P1570515ATMSSCIENCEAAT16032024518500.PDS",
        "type": "data"

        :param input_dict:
        :return:
        """
        # https://github.com/stac-extensions/file
        # https://github.com/stac-extensions/file/blob/main/examples/item.json
        description_keys = ['size', 'checksumType', 'checksum']
        descriptions = [f'{k}={input_dict[k]};' for k in description_keys if k in input_dict]
        asset = Asset(
            href=f"s3://{input_dict['bucket']}/{input_dict['key']}",
            title=input_dict['fileName'],
            description=''.join(descriptions),
            extra_fields={
                'file:size': input_dict['size'] if 'size' in input_dict else -1,
                'file:checksum': input_dict['checksum'] if 'checksum' in input_dict else '',
            },
            roles=[input_dict['type']]
        )
        return asset

    def to_stac(self, source: dict) -> dict:
        """
        Sample: Cumulus granule
        {
            "published": false,
            "endingDateTime": "2016-01-31T19:59:59.991043",
            "status": "completed",
            "timestamp": 1648050501578,
            "createdAt": 1648050499079,
            "processingEndDateTime": "2022-03-23T15:48:20.869Z",
            "productVolume": 18096656,
            "timeToPreprocess": 20.302,
            "timeToArchive": 0,
            "productionDateTime": "2016-02-01T02:45:59.639000Z",
            "execution": "https://console.aws.amazon.com/states/home?region=us-west-2#/executions/details/arn:aws:states:us-west-2:884500545225:execution:am-uds-dev-cumulus-IngestGranule:ec602ca7-0243-44df-adc0-28fb8a486d54",
            "files": [
                {
                    "bucket": "am-uds-dev-cumulus-internal",
                    "key": "ATMS_SCIENCE_Group___1/P1570515ATMSSCIENCEAAT16032024518500.PDS",
                    "size": 760,
                    "fileName": "P1570515ATMSSCIENCEAAT16032024518500.PDS",
                    "source": "data/SNPP_ATMS_Level0_T/ATMS_SCIENCE_Group/2016/031//P1570515ATMSSCIENCEAAT16032024518500.PDS",
                    "type": "data"
                },
                {
                    "bucket": "am-uds-dev-cumulus-internal",
                    "key": "ATMS_SCIENCE_Group___1/P1570515ATMSSCIENCEAAT16032024518501.PDS",
                    "size": 18084600,
                    "fileName": "P1570515ATMSSCIENCEAAT16032024518501.PDS",
                    "source": "data/SNPP_ATMS_Level0_T/ATMS_SCIENCE_Group/2016/031//P1570515ATMSSCIENCEAAT16032024518501.PDS",
                    "type": "metadata"
                },
                {
                    "bucket": "am-uds-dev-cumulus-internal",
                    "key": "ATMS_SCIENCE_Group___1/P1570515ATMSSCIENCEAAT16032024518501.PDS.xml",
                    "size": 9547,
                    "fileName": "P1570515ATMSSCIENCEAAT16032024518501.PDS.xml",
                    "source": "data/SNPP_ATMS_Level0_T/ATMS_SCIENCE_Group/2016/031//P1570515ATMSSCIENCEAAT16032024518501.PDS.xml",
                    "type": "metadata"
                },
                {
                    "bucket": "am-uds-dev-cumulus-internal",
                    "key": "ATMS_SCIENCE_Group___1/P1570515ATMSSCIENCEAAT16032024518500.PDS.cmr.xml",
                    "size": 1749,
                    "fileName": "P1570515ATMSSCIENCEAAT16032024518500.PDS.cmr.xml",
                    "type": "metadata"
                }
            ],
            "processingStartDateTime": "2022-03-23T15:45:03.732Z",
            "updatedAt": 1648050501578,
            "beginningDateTime": "2016-01-31T18:00:00.009057",
            "provider": "snpp_provider_03",
            "granuleId": "P1570515ATMSSCIENCEAAT16032024518500.PDS",
            "collectionId": "ATMS_SCIENCE_Group___001",
            "duration": 197.993,
            "error": {},
            "custom_metadata": {
                "custom_entry_1": "test1",
                "custom_entry_2": true,
                "custom_entry_3": 3.14159,
                "custom_entry_4": -20107,
            }
            "lastUpdateDateTime": "2018-04-25T21:45:45.524053"
        }
        :param source:
        :return:
        """

        validation_result = JsonValidator(self.__cumulus_granule_schema).validate(source)
        if validation_result is not None:
            raise ValueError(f'invalid cumulus granule json: {validation_result}')

        cumulus_file_validator = JsonValidator(CUMULUS_FILE_SCHEMA)
        validated_files = [k for k in source['files'] if cumulus_file_validator.validate(k) is None]
        custom_metadata = source['custom_metadata'] if 'custom_metadata' in source else {}

        properties_template = [
            {
                'stac_key': 'datetime',
                'cumulus_key': 'createdAt',
                'execution': lambda x: f"{TimeUtils.decode_datetime(x['createdAt'], False)}Z"
            },
            {
                'stac_key': 'start_datetime',
                'cumulus_key': 'beginningDateTime',
                'execution': lambda x: datetime_to_str(self.get_time_obj(x['beginningDateTime']))
            },
            {
                'stac_key': 'end_datetime',
                'cumulus_key': 'endingDateTime',
                'execution': lambda x: datetime_to_str(self.get_time_obj(x['endingDateTime']))
            },
            {
                'stac_key': 'created',
                'cumulus_key': 'productionDateTime',
                'execution': lambda x: datetime_to_str(self.get_time_obj(x['productionDateTime']))
            },
            {
                'stac_key': 'updated',
                'cumulus_key': 'updatedAt',
                'execution': lambda x: datetime_to_str(TimeUtils().parse_from_unix(source['updatedAt'], True).get_datetime_obj())
            },
            {
                'stac_key': 'status',
                'cumulus_key': 'status',
                'execution': lambda x: x['status']
            },
            {
                'stac_key': 'provider',
                'cumulus_key': 'provider',
                'execution': lambda x: x['provider']
            },
        ]
        cumulus_properties = {
            k['stac_key']: k['execution'](source) for k in properties_template if k['cumulus_key'] in source
        }
        stac_item = Item(
            id=source['granuleId'],
            stac_extensions=["https://stac-extensions.github.io/file/v2.1.0/schema.json"],
            bbox=[-180.0, -90.0, 180.0, 90.0],
            properties={
                **custom_metadata,
                **cumulus_properties,
                # "datetime": f"{TimeUtils.decode_datetime(source['createdAt'], False)}Z",
                # "start_datetime": datetime_to_str(self.get_time_obj(source['beginningDateTime'])),
                # "end_datetime": datetime_to_str(self.get_time_obj(source['endingDateTime'])),
                # "created": datetime_to_str(self.get_time_obj(source['productionDateTime'])),
                # "updated": datetime_to_str(TimeUtils().parse_from_unix(source['updatedAt'], True).get_datetime_obj()),
            },
            collection=source['collectionId'],
            assets={self.__get_asset_name(k): self.__get_asset_obj(k) for k in validated_files},
            geometry={
                "type": "Point",
                "coordinates": [0.0, 0.0]
            },
            datetime=TimeUtils().parse_from_unix(source['createdAt'], True).get_datetime_obj(),
        )
        stac_item.links = [
            Link(rel='collection', target='.')
        ]
        return stac_item.to_dict(include_self_link=False, transform_hrefs=False)

    def from_stac(self, source: dict) -> Item:
        return Item.from_dict(source)



from mdps_ds_lib.stage_in_out.search_collections_abstract import SearchCollectionsAbstract


class SearchCollectionsCmr(SearchCollectionsAbstract):
    def search(self, **kwargs):
        raise NotImplementedError()



class CatalogGranulesFactory:
    UNITY = 'UNITY'

    def get_class(self, upload_type):
        if upload_type == CatalogGranulesFactory.UNITY:
            from mdps_ds_lib.stage_in_out.catalog_granules_unity import CatalogGranulesUnity
            return CatalogGranulesUnity()
        raise ValueError(f'unknown search_type: {upload_type}')



import json

from mdps_ds_lib.lib.aws.aws_cred import AwsCred


class AwsLambda(AwsCred):
    def __init__(self):
        super().__init__()
        self.__lambda_client = self.get_client('lambda')

    def invoke_function(self, function_name: str, payload: dict):
        response = self.__lambda_client.invoke(
            FunctionName=function_name,
            InvocationType='Event',  # 'Event' = async | 'RequestResponse' = sync | 'DryRun',
            LogType='None',  # 'None' = async | 'Tail =  sync',
            ClientContext='',  # Up to 3583 bytes of base64-encoded data
            Payload=json.dumps(payload).encode(),
            # Qualifier='string'
        )
        return response



import json

from mdps_ds_lib.stage_in_out.stage_in_out_utils import StageInOutUtils

from mdps_ds_lib.stage_in_out.dapa_client import DapaClient
from mdps_ds_lib.lib.utils.time_utils import TimeUtils
from mdps_ds_lib.stage_in_out.catalog_granules_abstract import CatalogGranulesAbstract
import logging
import os

from mdps_ds_lib.stage_in_out.cataloging_granules_status_checker import CatalogingGranulesStatusChecker

LOGGER = logging.getLogger(__name__)


class CatalogGranulesUnity(CatalogGranulesAbstract):
    PROVIDER_ID_KEY = 'PROVIDER_ID'
    VERIFY_SSL_KEY = 'VERIFY_SSL'
    DELAY_SECOND = 'DELAY_SECOND'
    REPEAT_TIMES = 'REPEAT_TIMES'
    CHUNK_SIZE = 'CHUNK_SIZE'

    def __init__(self) -> None:
        super().__init__()
        self.__provider_id = ''
        self.__verify_ssl = True
        self.__delaying_second = 30
        self.__repeating_times = 0
        self.__chunk_size = StageInOutUtils.CATALOG_DEFAULT_CHUNK_SIZE

    def __set_props_from_env(self):
        missing_keys = [k for k in [self.UPLOADED_FILES_JSON, self.PROVIDER_ID_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        self._retrieve_stac_json()
        self.__chunk_size = int(os.environ.get(self.CHUNK_SIZE, StageInOutUtils.CATALOG_DEFAULT_CHUNK_SIZE))
        self.__chunk_size = self.__chunk_size if self.__chunk_size > 0 else StageInOutUtils.CATALOG_DEFAULT_CHUNK_SIZE
        self.__provider_id = os.environ.get(self.PROVIDER_ID_KEY)
        self.__verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        self.__delaying_second = int(os.environ.get(self.DELAY_SECOND, '30'))
        self.__repeating_times = int(os.environ.get(self.REPEAT_TIMES, '0'))
        self.__verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        return self

    def catalog(self, **kwargs):
        self.__set_props_from_env()
        if isinstance(self._uploaded_files_json, dict) and 'features' in self._uploaded_files_json:
            self._uploaded_files_json = self._uploaded_files_json['features']
        dapa_client = DapaClient().with_verify_ssl(self.__verify_ssl)
        main_collection = self._uploaded_files_json[0]['collection']
        collection_result = dapa_client.get_collection(main_collection)
        LOGGER.debug(f'{main_collection} details: {collection_result}')
        response_jsons = []
        for i, features_chunk in enumerate(StageInOutUtils.chunk_list(self._uploaded_files_json, self.__chunk_size)):
            LOGGER.debug(f'working on chunk_index {i}')
            dapa_body = {
                "provider_id": self.__provider_id,
                "features": features_chunk
            }
            LOGGER.debug(f'dapa_body_granules: {dapa_body}')
            dapa_ingest_result = dapa_client.ingest_granules_w_cnm(dapa_body)
            extracted_ids = [k['id'] for k in features_chunk]
            LOGGER.debug(f'checking following IDs: {extracted_ids}')
            status_checker = CatalogingGranulesStatusChecker(features_chunk[0]['collection'],
                                                             extracted_ids,
                                                             TimeUtils().get_datetime_obj().timestamp(),
                                                             self.__delaying_second,
                                                             self.__repeating_times,
                                                             self.__verify_ssl)
            status_result = status_checker.verify_n_times()
            LOGGER.debug(f'chunk_index {i} status_result: {status_result}')
            response_jsons.append({
                'cataloging_request_status': dapa_ingest_result,
                'status_result': status_result
            })
        return json.dumps(response_jsons)



import time

import requests

from mdps_ds_lib.lib.earthdata_login.urs_token_retriever import URSTokenRetriever
from mdps_ds_lib.stage_in_out.download_granules_abstract import DownloadGranulesAbstract
import logging
import os


LOGGER = logging.getLogger(__name__)


class DownloadGranulesDAAC(DownloadGranulesAbstract):

    def __init__(self) -> None:
        super().__init__()
        self.__edl_token = None
        self.__retry_wait_time_sec = int(os.environ.get('DOWNLOAD_RETRY_WAIT_TIME', '30'))
        self.__retry_times = int(os.environ.get('DOWNLOAD_RETRY_TIMES', '5'))

    def _set_props_from_env(self):
        missing_keys = [k for k in [self.STAC_JSON, self.DOWNLOAD_DIR_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        self._retrieve_stac_json()
        self._setup_download_dir()
        self.__edl_token = URSTokenRetriever().start()
        return self

    def _download_one_item(self, downloading_url):
        headers = {
            'Authorization': f'Bearer {self.__edl_token}'
        }
        r = requests.get(downloading_url, headers=headers)
        download_count = 1
        while r.status_code in [502, 504] and download_count < self.__retry_times:
            LOGGER.error(f'502 or 504 while downloading {downloading_url}. attempt: {download_count}')
            time.sleep(self.__retry_wait_time_sec)
            r = requests.get(downloading_url, headers=headers)
            download_count += 1
        r.raise_for_status()
        local_file_path = os.path.join(self._download_dir, os.path.basename(downloading_url))
        with open(local_file_path, 'wb') as fd:
            fd.write(r.content)
        return local_file_path



import logging
import os
from glob import glob

from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerAbstract, JobManagerProps
from mdps_ds_lib.lib.utils.file_utils import FileUtils

LOGGER = logging.getLogger(__name__)


class JobManagerLocalFileSystem(JobManagerAbstract):
    def __init__(self, props=JobManagerProps()) -> None:
        super().__init__()
        self.__props = props

    def get_all_job_files(self):
        return [k.replace(self.__props.job_bucket, '')[1:] for k in glob(os.path.join(self.__props.job_bucket, self.__props.job_path, f'*{self.__props.job_file_postfix}'))]

    def get_job_file(self, job_path, validate_job_content=lambda x: True):
        job_abs_path = os.path.join(self.__props.job_bucket, job_path)
        with self.__props.lock:
            try:
                job_content = FileUtils.read_json(job_abs_path)
            except:
                LOGGER.exception(f'job in this path is not in JSON: {job_path}')
                return None
            FileUtils.remove_if_exists(job_abs_path)
        if validate_job_content(job_content) is False:
            LOGGER.error(f'{job_path} does not have valid json: {job_content}. Putting back')
            FileUtils.write_json(job_abs_path, job_content, overwrite=True)
            return None
        LOGGER.debug(f'writing job to processing path')
        FileUtils.mk_dir_p(os.path.join(self.__props.job_bucket, self.__props.processing_job_path))
        FileUtils.write_json(os.path.join(self.__props.job_bucket, self.__props.processing_job_path, os.path.basename(job_path)), job_content, overwrite=True)
        return job_content

    def put_back_failed_job(self, original_job_path: str):
        processing_job_path = os.path.join(self.__props.processing_job_path, os.path.basename(original_job_path))
        LOGGER.debug(f'reading failed job file: {processing_job_path}')
        try:
            job_json: dict = FileUtils.read_json(os.path.join(self.__props.job_bucket, processing_job_path))
            LOGGER.debug(f'deleting failed job file from processing dir: {processing_job_path}')
            FileUtils.remove_if_exists(os.path.join(self.__props.job_bucket, processing_job_path))
            LOGGER.debug(f'uploading failed job back to folder')
            FileUtils.write_json(os.path.join(self.__props.job_bucket, original_job_path), job_json)
        except Exception as e:
            LOGGER.exception(f'error while reading failed job file from from processing dir: {processing_job_path}')
        return

    def remove_from_processing(self, job_path):
        FileUtils.remove_if_exists(os.path.join(self.__props.job_bucket, self.__props.processing_job_path, os.path.basename(job_path)))
        return True



import json
import logging
import time
from http.cookiejar import CookieJar
from typing import Dict
from urllib import request

import requests
from requests.auth import HTTPBasicAuth
from tenacity import retry, retry_if_result, stop_after_attempt, wait_random_exponential

LOGGER = logging.getLogger(__name__)

class URSToken(object):
    DNS_ERROR_TXT = 'Name-Resolution-Error'
    """
    Traceback (most recent call last):
 File "/usr/local/lib/python3.9/site-packages/urllib3/connection.py", line 174, in _new_conn
  conn = connection.create_connection(
 File "/usr/local/lib/python3.9/site-packages/urllib3/util/connection.py", line 72, in create_connection
  for res in socket.getaddrinfo(host, port, family, socket.SOCK_STREAM):
 File "/usr/local/lib/python3.9/socket.py", line 954, in getaddrinfo
  for res in _socket.getaddrinfo(host, port, family, type, proto, flags):
socket.gaierror: [Errno -3] Temporary failure in name resolution
During handling of the above exception, another exception occurred:
Traceback (most recent call last):
 File "/usr/local/lib/python3.9/site-packages/urllib3/connectionpool.py", line 714, in urlopen
  httplib_response = self._make_request(
 File "/usr/local/lib/python3.9/site-packages/urllib3/connectionpool.py", line 403, in _make_request
  self._validate_conn(conn)
 File "/usr/local/lib/python3.9/site-packages/urllib3/connectionpool.py", line 1053, in _validate_conn
  conn.connect()
 File "/usr/local/lib/python3.9/site-packages/urllib3/connection.py", line 363, in connect
  self.sock = conn = self._new_conn()
 File "/usr/local/lib/python3.9/site-packages/urllib3/connection.py", line 186, in _new_conn
  raise NewConnectionError(
urllib3.exceptions.NewConnectionError: <urllib3.connection.HTTPSConnection object at 0x7f111cbc54f0>: Failed to establish a new connection: [Errno -3] Temporary failure in name resolution
    """
    def __init__(self, username: str, dwssap: str, edl_base_url: str = None, wait_time = 30, retry_times = 5) -> None:
        super().__init__()
        self.__default_edl_base_url = 'https://urs.earthdata.nasa.gov/'
        self.__username = username
        self.__dwssap = dwssap
        self.__edl_base_url = self.__default_edl_base_url if edl_base_url is None or edl_base_url == '' else edl_base_url.strip().lower()
        if not self.__edl_base_url.endswith('/'):
            self.__edl_base_url = f'{self.__edl_base_url}/'
        if not self.__edl_base_url.startswith('http'):
            self.__edl_base_url = f'https://{self.__edl_base_url}'
        self.__token = None
        self.__wait_time = wait_time
        self.__retry_times = retry_times


    @retry(wait=wait_random_exponential(multiplier=1, max=60),
           stop=stop_after_attempt(3),
           reraise=True,
           retry=(retry_if_result(lambda x: x == ''))
           )
    def create_token(self, url: str) -> str:
        token: str = ''
        try:
            headers: Dict = {'Accept': 'application/json'}  # noqa E501
            resp = requests.post(url + "/token", headers=headers, auth=HTTPBasicAuth(self.__username, self.__dwssap))
            if resp.status_code > 300:
                raise ValueError(f'invalid response code: {resp.status_code}. details: {resp.content}')
            response_content: Dict = json.loads(resp.content)
            if "error" in response_content:
                if response_content["error"] == "max_token_limit":
                    LOGGER.error("Max tokens acquired from URS. Using existing token")
                    tokens = self.list_tokens(url)
                    return tokens[0]
            if 'access_token' not in response_content:
                raise ValueError(f'access_token not found. {response_content}')
            token = response_content['access_token']

        # Add better error handling there
        # Max tokens
        # Wrong Username/Password
        # Other
        except Exception as e:  # noqa E722
            LOGGER.warning("Error getting the token - check user name and password", exc_info=True)
            raise RuntimeError(str(e))
        return token

    def list_tokens(self, url: str):
        tokens = []
        try:
            headers: Dict = {'Accept': 'application/json'}  # noqa E501
            resp = requests.get(url + "/tokens", headers=headers, auth=HTTPBasicAuth(self.__username, self.__dwssap))
            if resp.status_code >= 400:
                LOGGER.error(f'error response: {resp.status_code}. details: {resp.content}')
                return tokens
            response_content = json.loads(resp.content)

            for x in response_content:
                tokens.append(x['access_token'])

        except Exception as e:  # noqa E722
            LOGGER.warning("Error getting the token - check user name and password", exc_info=True)
        return tokens

    def setup_earthdata_login_auth(self):
        """
        Set up the request library so that it authenticates against the given
        Earthdata Login endpoint and is able to track cookies between requests.
        This looks in the .netrc file first and if no credentials are found,
        it prompts for them.
        Valid endpoints include:
            urs.earthdata.nasa.gov - Earthdata Login production
        """
        manager = request.HTTPPasswordMgrWithDefaultRealm()
        manager.add_password(None, self.__edl_base_url, self.__username, self.__dwssap)
        auth = request.HTTPBasicAuthHandler(manager)
        __version__ = "1.12.0"

        jar = CookieJar()
        processor = request.HTTPCookieProcessor(jar)
        opener = request.build_opener(auth, processor)
        opener.addheaders = [('User-agent', 'unity-downloader-' + __version__)]
        request.install_opener(opener)

    def delete_token(self, token: str) -> bool:
        try:
            self.setup_earthdata_login_auth()
            headers: Dict = {'Accept': 'application/json'}
            resp = requests.post(f'{self.__edl_base_url}revoke_token', params={"token": token}, headers=headers,
                                 auth=HTTPBasicAuth(self.__username, self.__dwssap))

            if resp.status_code == 200:
                LOGGER.info("EDL token successfully deleted")
                return True
            else:
                LOGGER.info("EDL token deleting failed.")

        except:  # noqa E722
            LOGGER.warning("Error deleting the token", exc_info=True)
        return False

    def __get_token_once(self):
        try:
            token_url = f'{self.__edl_base_url}api/users'
            tokens = self.list_tokens(token_url)
            if len(tokens) == 0:
                return self.create_token(token_url)
        except Exception as conn_err:
            if 'Temporary failure in name resolution' in str(conn_err):
                return self.DNS_ERROR_TXT
            raise conn_err
        return tokens[0]

    def get_token(self) -> str:
        token = self.__get_token_once()
        retry_count = 0
        while token == self.DNS_ERROR_TXT and retry_count < self.__retry_times:
            LOGGER.error(f'{self.DNS_ERROR_TXT} for URS Token. attempt: {retry_count}')
            time.sleep(self.__wait_time)
            token = self.__get_token_once()
            retry_count += 1
        return token




from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerAbstract, JobManagerProps


class JobManagerMemory(JobManagerAbstract):
    def __init__(self, props=JobManagerProps()) -> None:
        self.__props = props
        self.__processing_job_dict = {}

    def get_all_job_files(self):
        return [k for k in self.__props.memory_job_dict.keys()]

    def get_job_file(self, job_path, validate_job_content=lambda x: True):
        if job_path not in self.__props.memory_job_dict:
            return None
        job = self.__props.memory_job_dict.get(job_path)
        del self.__props.memory_job_dict[job_path]
        if not validate_job_content(job):
            return None
        self.__processing_job_dict[job_path] = job
        return job

    def remove_from_processing(self, job_path):
        if job_path not in self.__processing_job_dict:
            return
        del self.__processing_job_dict[job_path]
        pass

    def put_back_failed_job(self, original_job_path: str):
        if original_job_path not in self.__processing_job_dict:
            return
        self.__props.memory_job_dict[original_job_path] = self.__processing_job_dict[original_job_path]
        del self.__processing_job_dict[original_job_path]
        return



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import os
import re
import zlib
from functools import partial
from pathlib import Path
from subprocess import Popen, PIPE
from distutils.dir_util import copy_tree


class FileUtils:
    @staticmethod
    def remove_if_exists(file_path):
        if os.path.exists(file_path) and os.path.isfile(file_path):
            os.remove(file_path)
        return

    @staticmethod
    def is_relative_path(file_path):
        regex = re.compile('^([A-Za-z0-9-]+)://.+$')
        current_granule_id = re.findall(regex, file_path)
        if len(current_granule_id) > 0:
            return False
        return not os.path.isabs(file_path)

    @staticmethod
    def mk_dir_p(dir_path):
        Path(dir_path).mkdir(parents=True, exist_ok=True)
        return

    @staticmethod
    def copy_dir(source_dir, dest_dir):
        copy_tree(source_dir, dest_dir)
        return

    @staticmethod
    def gunzip_file_os(zipped_file_path, output_file_path=None):
        if not FileUtils.file_exist(zipped_file_path):
            raise ValueError('missing file: {}'.format(zipped_file_path))
        session = Popen(['gunzip', zipped_file_path], stdout=PIPE, stderr=PIPE)
        stdout, stderr = session.communicate()
        if stderr:
            raise RuntimeError('error while gunzipping the file with Popen. filename: {}. error: {}'.format(zipped_file_path, stderr))
        default_output_path = zipped_file_path[:-3]
        if not FileUtils.file_exist(default_output_path):
            raise ValueError('missing gunzipped file: {}'.format(default_output_path))
        if output_file_path is None:
            output_file_path = default_output_path
        if FileUtils.file_exist(output_file_path) and default_output_path != output_file_path:
            os.renames(default_output_path, output_file_path)
        return output_file_path

    @staticmethod
    def get_checksum(file_path, is_md5=False, chunk_size=25 * 2**20):
        with open(file_path, mode='rb') as f:
            d = hashlib.md5() if is_md5 else hashlib.sha512()
            for buf in iter(partial(f.read, chunk_size), b''):
                d.update(buf)
        return d.hexdigest()

    @staticmethod
    def get_size(file_path):
        return os.stat(file_path).st_size

    @staticmethod
    def file_exist(path):
        return Path(path).is_file()

    @staticmethod
    def dir_exist(path):
        return Path(path).is_dir()

    @staticmethod
    def del_file(path):
        if FileUtils.file_exist(path):
            Path(path).unlink()
        return

    @staticmethod
    def read_json(path):
        if not os.path.exists(path):
            raise IOError('{} does not exist'.format(path))
        with open(path, 'r') as ff:
            try:
                return json.loads(ff.read())
            except:
                return None

    @staticmethod
    def write_json(file_path, json_obj, overwrite=False, append=False, prettify=False):
        if os.path.exists(file_path) and not overwrite:
            raise ValueError('{} already exists, and not overwriting'.format(file_path))
        with open(file_path, 'a' if append else 'w') as ff:
            json_str = json.dumps(json_obj, indent=4) if prettify else json.dumps(json_obj)
            ff.write(json_str)
        return



import logging
import os
from io import BytesIO
from typing import Union

from mdps_ds_lib.lib.aws.aws_cred import AwsCred
from mdps_ds_lib.lib.utils.file_utils import FileUtils

LOGGER = logging.getLogger(__name__)


class AwsS3(AwsCred):
    def __init__(self):
        super().__init__()
        self.__valid_s3_schemas = ['s3://', 's3a://', 's3s://']
        self.__s3_client = self.get_client('s3')
        self.__s3_resource = self.get_resource('s3')
        self.__target_bucket = None
        self.__target_key = None

    def __upload_to_s3(self, bucket, prefix, file_path, delete_files=False, add_size=True, other_tags={}, s3_name=None):
        """
        Uploading a file to S3
        :param bucket: string - name of bucket
        :param prefix: string - prefix. don't start and end with `/` to avoid extra unnamed dirs
        :param file_path: string - absolute path of file location
        :param delete_files: boolean - deleting original file. default: False
        :param add_size: boolean - adding the file size as tag. default: True
        :param other_tags: dict - key-value pairs as a dictionary
        :param s3_name: string - name of s3 file if the user wishes to change.
                    using the actual filename if not provided. defaulted to None
        :return: None
        """
        tags = {
            'TagSet': []
        }
        if add_size is True:
            tags['TagSet'].append({
                'Key': 'org_size',
                'Value': str(FileUtils.get_size(file_path))
            })
        for key, val in other_tags.items():
            tags['TagSet'].append({
                'Key': key,
                'Value': str(val)
            })
        if s3_name is None:
            s3_name = os.path.basename(file_path)
        s3_key = '{}/{}'.format(prefix, s3_name)
        self.__s3_client.upload_file(file_path, bucket, s3_key, ExtraArgs={'ServerSideEncryption': 'AES256'})
        if delete_files is True:  # deleting local files
            FileUtils.remove_if_exists(file_path)
        if len(tags['TagSet']) > 0:
            try:
                self.__s3_client.put_object_tagging(Bucket=bucket, Key=s3_key, Tagging=tags)
            except Exception as e:
                LOGGER.exception(f'error while adding tags: {tags} to {bucket}/{s3_key}')
                raise e
        return f's3://{bucket}/{s3_key}'

    def exists(self, base_path: str, relative_path: str):
        try:
            response = self.__s3_client.head_object(Bucket=base_path, Key=relative_path)
        except:
            return False
        return True

    def upload(self, file_path: str, base_path: str, relative_parent_path: str, delete_files: bool,
               s3_name: Union[str, None] = None, obj_tags: dict = {}, overwrite: bool = False):
        s3_url = self.__upload_to_s3(base_path, relative_parent_path, file_path, delete_files, True, obj_tags, s3_name)
        if delete_files is True:  # deleting local files
            FileUtils.remove_if_exists(file_path)
        return s3_url

    def get_s3_stream(self):
        return self.__s3_client.get_object(Bucket=self.__target_bucket, Key=self.__target_key)['Body']

    def get_s3_obj_size(self):
        # get head of the s3 file
        s3_obj_head = self.__s3_client.head_object(
            Bucket=self.__target_bucket,
            Key=self.__target_key,
        )
        # get the object size
        s3_obj_size = int(s3_obj_head['ResponseMetadata']['HTTPHeaders']['content-length'])
        if s3_obj_size is None:  # no object size found. something went wrong.
            return -1
        return s3_obj_size

    def __get_all_s3_files_under(self, bucket, prefix, with_versions=False):
        list_method_name = 'list_object_versions' if with_versions is True else 'list_objects_v2'
        page_key = 'Versions' if with_versions is True else 'Contents'
        paginator = self.__s3_client.get_paginator(list_method_name)
        operation_parameters = {
            'Bucket': bucket,
            'Prefix': prefix
        }
        page_iterator = paginator.paginate(**operation_parameters)
        for eachPage in page_iterator:
            if page_key not in eachPage:
                continue
            for fileObj in eachPage[page_key]:
                yield fileObj

    def get_child_s3_files(self, bucket, prefix, additional_checks=lambda x: True, with_versions=False):
        for fileObj in self.__get_all_s3_files_under(bucket, prefix, with_versions=with_versions):
            if additional_checks(fileObj):
                yield fileObj['Key'], fileObj['Size']

    def set_s3_url(self, s3_url):
        LOGGER.debug(f'setting s3_url: {s3_url}')
        self.__target_bucket, self.__target_key = self.split_s3_url(s3_url)
        LOGGER.debug(f'props: {self.__target_bucket}, {self.__target_key}')
        return self

    def split_s3_url(self, s3_url):
        s3_schema = [k for k in self.__valid_s3_schemas if s3_url.startswith(k)]
        if len(s3_schema) != 1:
            raise ValueError('invalid s3 url: {}'.format(s3_url))

        s3_schema_length = len(s3_schema[0])
        split_index = s3_url[s3_schema_length:].find('/')
        bucket = s3_url[s3_schema_length: split_index + s3_schema_length]
        key = s3_url[(split_index + s3_schema_length + 1):]
        return bucket, key

    def __tag_existing_obj(self, other_tags={}):
        if len(other_tags) == 0:
            return
        tags = {
            'TagSet': []
        }
        for key, val in other_tags.items():
            tags['TagSet'].append({
                'Key': key,
                'Value': str(val)
            })
        self.__s3_client.put_object_tagging(Bucket=self.__target_bucket, Key=self.__target_key, Tagging=tags)
        return

    def add_tags_to_obj(self, other_tags={}):
        """
        retrieve existing tags first and append new tags to them

        :param bucket: string
        :param s3_key: string
        :param other_tags: dict
        :return: bool
        """
        if len(other_tags) == 0:
            return False
        response = self.__s3_client.get_object_tagging(Bucket=self.__target_bucket, Key=self.__target_key)
        if 'TagSet' not in response:
            return False
        all_tags = {k['Key']: k['Value'] for k in response['TagSet']}
        for k, v in other_tags.items():
            all_tags[k] = v
            pass
        self.__tag_existing_obj(all_tags)
        return True

    def download(self, local_dir, file_name=None):

        if not FileUtils.dir_exist(local_dir):
            raise ValueError('missing directory')
        if file_name is None:
            LOGGER.debug(f'setting the downloading filename from target_key: {self.__target_key}')
            file_name = os.path.basename(self.__target_key)
        local_file_path = os.path.join(local_dir, file_name)
        LOGGER.debug(f'downloading to local_file_path: {local_file_path}')
        self.__s3_client.download_file(self.__target_bucket, self.__target_key, local_file_path)
        LOGGER.debug(f'file downloaded')
        return local_file_path

    @property
    def target_bucket(self):
        return self.__target_bucket

    @target_bucket.setter
    def target_bucket(self, val):
        """
        :param val:
        :return: None
        """
        self.__target_bucket = val
        return

    @property
    def target_key(self):
        return self.__target_key

    @target_key.setter
    def target_key(self, val):
        """
        :param val:
        :return: None
        """
        self.__target_key = val
        return

    def get_size(self):
        response = self.__s3_client.head_object(Bucket=self.target_bucket, Key=self.target_key)
        return response['ContentLength']

    def get_stream(self):
        if self.target_bucket is None or self.target_key is None:
            raise ValueError('bucket or key is None. Set them before calling this method')
        return self.__s3_client.get_object(Bucket=self.target_bucket, Key=self.target_key)['Body']

    def upload_bytes(self, content: bytes):
        self.__s3_client.put_object(Bucket=self.target_bucket,
                                    Key=self.target_key,
                                    ContentType='binary/octet-stream',
                                    Body=content,
                                    ServerSideEncryption='AES256')
        return

    def read_small_txt_file(self):
        """
        convenient method to read small text files stored in S3
        :return: text file contents
        """
        bytestream = BytesIO(self.get_stream().read())  # get the bytes stream of zipped file
        return bytestream.read().decode('UTF-8')



import json
import logging
import os


from mdps_ds_lib.lib.constants import Constants
from mdps_ds_lib.stage_in_out.dapa_client import DapaClient
from mdps_ds_lib.stage_in_out.search_granules_abstract import SearchGranulesAbstract

LOGGER = logging.getLogger(__name__)


class SearchGranulesUnity(SearchGranulesAbstract):
    COLLECTION_ID_KEY = 'COLLECTION_ID'
    DOWNLOAD_DIR_KEY = 'DOWNLOAD_DIR'

    LIMITS_KEY = 'LIMITS'
    DATE_FROM_KEY = 'DATE_FROM'
    DATE_TO_KEY = 'DATE_TO'
    VERIFY_SSL_KEY = 'VERIFY_SSL'

    def __init__(self) -> None:
        super().__init__()
        self.__collection_id = ''
        self.__date_from = ''
        self.__date_to = ''
        self.__limit = 1000
        self.__verify_ssl = True
        self.__api_base_prefix = Constants.API_PREFIX

    def __set_props_from_env(self):
        missing_keys = [k for k in [self.COLLECTION_ID_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')

        self.__collection_id = os.environ.get(self.COLLECTION_ID_KEY)
        if self.LIMITS_KEY not in os.environ:
            LOGGER.warning(f'missing {self.LIMITS_KEY}. using default: {self.__limit}')
        else:
            self.__limit = int(os.environ.get(self.LIMITS_KEY))

        self.__date_from = os.environ.get(self.DATE_FROM_KEY, '')
        self.__date_to = os.environ.get(self.DATE_TO_KEY, '')
        self.__api_base_prefix = os.environ.get('API_PREFIX', Constants.API_PREFIX)
        self.__verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        return self

    def search(self, **kwargs) -> str:
        self.__set_props_from_env()
        dapa_client = DapaClient(self.__api_base_prefix).with_verify_ssl(self.__verify_ssl)
        granules_result = dapa_client.get_all_granules(self.__collection_id, self.__limit, self.__date_from, self.__date_to)
        granules_result = granules_result
        return json.dumps(granules_result)



from mdps_ds_lib.lib.aws.aws_cred import AwsCred


class AwsSns(AwsCred):
    def __init__(self):
        super().__init__()
        self.__sns_client = self.get_client('sns')
        self.__topic_arn = ''

    def set_topic_arn(self, topic_arn):
        self.__topic_arn = topic_arn
        return self

    def publish_message(self, msg_str: str):
        if self.__topic_arn == '':
            raise ValueError('missing topic arn to publish message')
        response = self.__sns_client.publish(
            TopicArn=self.__topic_arn,
            # TargetArn='string',  # not needed coz of we are using topic arn
            # PhoneNumber='string',  # not needed coz of we are using topic arn
            Message=msg_str,
            # Subject='optional string',
            # MessageStructure='string',
            # MessageAttributes={
            #     'string': {
            #         'DataType': 'string',
            #         'StringValue': 'string',
            #         'BinaryValue': b'bytes'
            #     }
            # },
            # MessageDeduplicationId='string',
            # MessageGroupId='string'
        )
        return response

    def create_sqs_subscription(self, sqs_arn):
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sns/client/subscribe.html
        if self.__topic_arn == '':
            raise ValueError('missing topic arn to publish message')
        response = self.__sns_client.subscribe(
            TopicArn=self.__topic_arn,
            Protocol='sqs',
            Endpoint=sqs_arn,  # For the sqs protocol, the endpoint is the ARN of an Amazon SQS queue.
            # Attributes={
            #     'string': 'string'
            # },
            ReturnSubscriptionArn=True  # if the API request parameter ReturnSubscriptionArn is true, then the value is always the subscription ARN, even if the subscription requires confirmation.
        )
        return response['SubscriptionArn']



from abc import ABCMeta, abstractmethod

from mdps_ds_lib.lib.utils.fake_lock import FakeLock


class JobManagerProps:
    def __init__(self):
        self.__job_bucket = None
        self.__job_path = None
        self.__job_file_postfix = ''
        self.__processing_job_path = None
        self.__lock = FakeLock()
        self.__memory_job_dict = {}

    @property
    def memory_job_dict(self):
        return self.__memory_job_dict

    @memory_job_dict.setter
    def memory_job_dict(self, val):
        """
        :param val:
        :return: None
        """
        self.__memory_job_dict = val
        return

    def load_from_json(self, input_json: dict):
        if 'memory_job_dict' in input_json:
            self.memory_job_dict = input_json['memory_job_dict']
        if 'job_bucket' in input_json:
            self.job_bucket = input_json['job_bucket']
        if 'job_path' in input_json:
            self.job_path = input_json['job_path']
        if 'processing_job_path' in input_json:
            self.processing_job_path = input_json['processing_job_path']
        if 'job_file_postfix' in input_json:
            self.job_file_postfix = input_json['job_file_postfix']
        return self

    @property
    def job_file_postfix(self):
        return self.__job_file_postfix

    @job_file_postfix.setter
    def job_file_postfix(self, val):
        """
        :param val:
        :return: None
        """
        self.__job_file_postfix = val
        return

    @property
    def lock(self):
        return self.__lock

    @lock.setter
    def lock(self, val):
        """
        :param val:
        :return: None
        """
        self.__lock = val
        return

    @property
    def job_bucket(self):
        return self.__job_bucket

    @job_bucket.setter
    def job_bucket(self, val):
        """
        :param val:
        :return: None
        """
        self.__job_bucket = val
        return

    @property
    def job_path(self):
        return self.__job_path

    @job_path.setter
    def job_path(self, val):
        """
        :param val:
        :return: None
        """
        self.__job_path = val
        return

    @property
    def processing_job_path(self):
        return self.__processing_job_path

    @processing_job_path.setter
    def processing_job_path(self, val):
        """
        :param val:
        :return: None
        """
        self.__processing_job_path = val
        return


class JobManagerAbstract(metaclass=ABCMeta):
    @abstractmethod
    def get_all_job_files(self):
        return

    @abstractmethod
    def get_job_file(self, job_path, validate_job_content=lambda x: True):
        return

    @abstractmethod
    def remove_from_processing(self, job_path):
        return

    @abstractmethod
    def put_back_failed_job(self, original_job_path: str):
        return


import json
import logging
import os.path
from glob import glob
from multiprocessing import Manager

from mdps_ds_lib.lib.cumulus_stac.granules_catalog import GranulesCatalog

from mdps_ds_lib.lib.aws.aws_s3 import AwsS3
from mdps_ds_lib.lib.processing_jobs.job_manager_memory import JobManagerMemory
from mdps_ds_lib.lib.processing_jobs.multithread_processor import MultiThreadProcessorProps, MultiThreadProcessor
from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerProps
from mdps_ds_lib.lib.utils.file_utils import FileUtils
from mdps_ds_lib.lib.processing_jobs.job_executor_abstract import JobExecutorAbstract
from mdps_ds_lib.lib.utils.time_utils import TimeUtils
from mdps_ds_lib.stage_in_out.upload_granules_abstract import UploadGranulesAbstract
from pystac import Item, Asset, ItemCollection, Catalog, Link

LOGGER = logging.getLogger(__name__)


class UploadItemExecutor(JobExecutorAbstract):
    def __init__(self, result_list, error_list, collection_id, staging_bucket, retry_wait_time_sec, retry_times, delete_files: bool) -> None:
        super().__init__()
        self.__collection_id = collection_id
        self.__staging_bucket = staging_bucket
        self.__delete_files = delete_files

        self.__gc = GranulesCatalog()
        self.__result_list = result_list
        self.__error_list = error_list
        # self.__gc = GranulesCatalog()
        self.__s3 = AwsS3()
        self.__retry_wait_time_sec = retry_wait_time_sec
        self.__retry_times = retry_times

    def validate_job(self, job_obj):
        return True

    def generate_sample_stac(self, filepath: str):
        filename = os.path.basename(filepath)
        file_checksum = FileUtils.get_checksum(filepath, True)
        # https://github.com/stac-extensions/file
        # https://github.com/stac-extensions/file/blob/main/examples/item.json
        sample_stac_item = Item(
                         id=f'{self.__collection_id}:{os.path.splitext(filename)[0]}',
                         stac_extensions=["https://stac-extensions.github.io/file/v2.1.0/schema.json"],
                         geometry={
                             "type": "Point",
                             "coordinates": [0.0, 0.0]
                         },
                         bbox=[0.0, 0.0, 0.0, 0.0],
                         datetime=TimeUtils().parse_from_unix(0, True).get_datetime_obj(),
                         properties={
                             "start_datetime": TimeUtils.get_current_time(),
                             "end_datetime": TimeUtils.get_current_time(),
                             "created": TimeUtils.get_current_time(),
                             "updated": TimeUtils.get_current_time(),
                         },
                         collection=self.__collection_id,
                         assets={
                             filename: Asset(
                                 href=filepath,
                                 roles=['data'],
                                 title=os.path.basename(filename),
                                 extra_fields={
                                     'file:size': FileUtils.get_size(filepath),
                                     'file:checksum': file_checksum,
                                 },
                                 description=f'size={FileUtils.get_size(filepath)};checksumType=md5;checksum={file_checksum}'),
                             f'{filename}.stac.json': Asset(href=f'{filepath}.stac.json', roles=['metadata'], description='desc=metadata stac;size=-1;checksumType=md5;checksum=unknown'),  # How to update this? It's a circular dependency
                         })

        return sample_stac_item

    def execute_job(self, job_obj, lock) -> bool:
        sample_stac_item = self.generate_sample_stac(job_obj)
        updating_assets = {}
        try:
            s3_url = self.__s3.upload(job_obj, self.__staging_bucket, f'{self.__collection_id}/{self.__collection_id}:{sample_stac_item.id}', self.__delete_files)
            updating_assets[os.path.basename(s3_url)] = s3_url
            uploading_current_granule_stac = f'{s3_url}.stac.json'
            self.__s3.set_s3_url(uploading_current_granule_stac)
            self.__s3.upload_bytes(json.dumps(sample_stac_item.to_dict(False, False),indent=4).encode())
            updating_assets[os.path.basename(uploading_current_granule_stac)] = uploading_current_granule_stac
            self.__gc.update_assets_href(sample_stac_item, updating_assets)
            self.__result_list.put(sample_stac_item.to_dict(False, False))
        except Exception as e:
            sample_stac_item.properties['upload_error'] = str(e)
            LOGGER.exception(f'error while processing: {job_obj}')
            self.__error_list.put(sample_stac_item.to_dict(False, False))
        return True


class UploadArbitraryFilesAsGranules(UploadGranulesAbstract):
    BASE_DIRECTORY = 'BASE_DIRECTORY'

    def __init__(self):
        super().__init__()
        self.__s3 = AwsS3()

    def upload(self, **kwargs) -> str:

        """
        1. Use Glob to find files
        2. Create stac.json for each file.
        3. Need collection ID which has tenant + venue.
        4. Create successful features.json
        :param kwargs:
        :return:
        """
        self._set_props_from_env()
        output_dir = os.environ.get(self.OUTPUT_DIRECTORY)
        if not FileUtils.dir_exist(output_dir):
            raise ValueError(f'OUTPUT_DIRECTORY: {output_dir} does not exist')
        missing_keys = [k for k in [self.BASE_DIRECTORY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')
        base_dir = os.environ.get(self.BASE_DIRECTORY)
        possible_files = [k for k in glob(os.path.join(base_dir, '**'), recursive=True) if os.path.isfile(k)]

        local_items = Manager().Queue()
        error_list = Manager().Queue()

        if self._parallel_count == 1:
            for each_child in possible_files:
                temp_job = UploadItemExecutor(local_items, error_list, self._collection_id, self._staging_bucket, self._retry_wait_time_sec, self._retry_times, self._delete_files)
                temp_job.execute_job(each_child, None)
        else:
            job_manager_props = JobManagerProps()
            for each_child in possible_files:
                job_manager_props.memory_job_dict[each_child] = each_child
            # https://www.infoworld.com/article/3542595/6-python-libraries-for-parallel-processing.html
            multithread_processor_props = MultiThreadProcessorProps(self._parallel_count)
            multithread_processor_props.job_manager = JobManagerMemory(job_manager_props)
            multithread_processor_props.job_executor = UploadItemExecutor(local_items, error_list, self._collection_id, self._staging_bucket, self._retry_wait_time_sec, self._retry_times, self._delete_files)
            multithread_processor = MultiThreadProcessor(multithread_processor_props)
            multithread_processor.start()

        LOGGER.debug(f'finished uploading all granules')
        dapa_body_granules = []
        while not local_items.empty():
            dapa_body_granules.append(local_items.get())

        errors = []
        while not error_list.empty():
            errors.append(error_list.get())
        LOGGER.debug(f'successful count: {len(dapa_body_granules)}. failed count: {len(errors)}')
        successful_item_collections = ItemCollection(items=dapa_body_granules)
        failed_item_collections = ItemCollection(items=errors)
        successful_features_file = os.path.join(output_dir, 'successful_features.json')

        failed_features_file = os.path.join(output_dir, 'failed_features.json')
        LOGGER.debug(f'writing results: {successful_features_file} && {failed_features_file}')
        FileUtils.write_json(successful_features_file, successful_item_collections.to_dict(False))
        FileUtils.write_json(failed_features_file, failed_item_collections.to_dict(False))
        s3_url = self.__s3.upload(successful_features_file, self._staging_bucket,
                                  self._result_path_prefix,
                                  s3_name=f'successful_features_{TimeUtils.get_current_time()}.json',
                                  delete_files=self._delete_files)
        LOGGER.debug(f'uploaded successful features to S3: {s3_url}')
        LOGGER.debug(f'creating response catalog')
        catalog = Catalog(
            id='NA',
            description='NA')
        catalog.add_link(Link('item', successful_features_file, 'application/json'))
        catalog.add_link(Link('item', failed_features_file, 'application/json'))
        catalog_json = catalog.to_dict(False, False)
        LOGGER.debug(f'catalog_json: {catalog_json}')
        return json.dumps(catalog_json)



from abc import ABC, abstractmethod
from typing import Any, Union, Callable

DEFAULT_TYPE = '_doc'


class ESAbstract(ABC):
    @abstractmethod
    def migrate_index_data(self, old_index, new_index):
        return
    @abstractmethod
    def create_index(self, index_name, index_body):
        return

    @abstractmethod
    def get_index_mapping(self, index_name):
        return {}

    @abstractmethod
    def has_index(self, index_name):
        return

    @abstractmethod
    def swap_index_for_alias(self, alias_name, old_index_name, new_index_name):
        return

    @abstractmethod
    def get_alias(self, alias_name):
        return

    @abstractmethod
    def create_alias(self, index_name, alias_name):
        return

    @abstractmethod
    def delete_index(self, index_name):
        return

    @abstractmethod
    def index_many(self, docs=None, doc_ids=None, doc_dict=None, index=None):
        return

    @abstractmethod
    def index_one(self, doc, doc_id, index=None):
        return

    @abstractmethod
    def update_many(self, docs=None, doc_ids=None, doc_dict=None, index=None):
        return

    @abstractmethod
    def update_one(self, doc, doc_id, index=None):
        return

    @staticmethod
    @abstractmethod
    def get_result_size(result):
        return

    @abstractmethod
    def query_with_scroll(self, dsl, querying_index=None):
        return

    @abstractmethod
    def query(self, dsl, querying_index=None):
        return

    @abstractmethod
    def delete_by_query(self, dsl, querying_index=None):
        return

    @abstractmethod
    def query_pages(self, dsl, querying_index=None):
        return

    @abstractmethod
    def query_by_id(self, doc_id, querying_index=None):
        return



import json
import logging
import os

import requests

from mdps_ds_lib.stage_in_out.search_granules_abstract import SearchGranulesAbstract

LOGGER = logging.getLogger(__name__)


class SearchGranulesCmr(SearchGranulesAbstract):
    CMR_BASE_URL_KEY = 'CMR_BASE_URL'
    COLLECTION_ID_KEY = 'COLLECTION_ID'
    DOWNLOAD_DIR_KEY = 'DOWNLOAD_DIR'

    LIMITS_KEY = 'LIMITS'
    DATE_FROM_KEY = 'DATE_FROM'
    DATE_TO_KEY = 'DATE_TO'
    VERIFY_SSL_KEY = 'VERIFY_SSL'

    def __init__(self) -> None:
        super().__init__()
        self.__collection_id = ''
        self.__date_from = ''
        self.__date_to = ''
        self.__limit = 1000
        self.__page_size = 2000  # page_size - number of results per page - default is 10, max is 2000
        self.__verify_ssl = True
        self.__cmr_base_url = ''

    def __set_props_from_env(self):
        missing_keys = [k for k in [self.COLLECTION_ID_KEY, self.CMR_BASE_URL_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')

        self.__collection_id = os.environ.get(self.COLLECTION_ID_KEY)
        self.__cmr_base_url = os.environ.get(self.CMR_BASE_URL_KEY)
        if not self.__cmr_base_url.endswith('/'):
            self.__cmr_base_url = f'{self.__cmr_base_url}/'
        if self.LIMITS_KEY not in os.environ:
            LOGGER.warning(f'missing {self.LIMITS_KEY}. using default: {self.__limit}')
        else:
            self.__limit = int(os.environ.get(self.LIMITS_KEY))

        self.__date_from = os.environ.get(self.DATE_FROM_KEY, '')
        self.__date_to = os.environ.get(self.DATE_TO_KEY, '')
        self.__verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        return self

    def __get_correct_result_count(self, results):
        if self.__limit < 0 or self.__limit >= len(results):
            return results
        return results[0: self.__limit]

    def search(self, **kwargs) -> str:
        """
  curl 'https://cmr.earthdata.nasa.gov/search/granules.stac' \
  -H 'accept: application/json; profile=stac-catalogue' \
  -H 'content-type: application/x-www-form-urlencoded' \
  --data-raw 'collection_concept_id=C1649553296-PODAAC&page_num=1&page_size=20&temporal[]=2011-08-01T00:00:00,2011-09-01T00:00:00'

        https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#stac
        https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#query-parameters

        :param kwargs:
        :return:
        """
        self.__set_props_from_env()
        header = {
            'accept': 'application/json; profile=stac-catalogue',
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        results = []
        page_size = self.__page_size if self.__limit < 0 or self.__limit > self.__page_size else self.__limit
        page_num = 1
        items_collection_shell = None
        while True:
            if 0 < self.__limit <= len(results):
                break
            request_body = {
                'collection_concept_id': self.__collection_id,
                'page_num': str(page_num),
                'page_size': str(page_size),
                'temporal[]': f'{self.__date_from},{self.__date_to}'
            }
            cmr_granules_url = f'{self.__cmr_base_url}search/granules.stac'
            response = requests.post(url=cmr_granules_url, headers=header, verify=self.__verify_ssl,
                                     data=request_body)
            if response.status_code >= 400:
                raise RuntimeError(
                    f'Cognito ends in error. status_code: {response.status_code}. url: {cmr_granules_url}. details: {response.text}')
            temp_results = json.loads(response.content.decode('utf-8'))
            if items_collection_shell is None:
                items_collection_shell = temp_results
            temp_results = temp_results.pop('features')
            page_num += 1
            results.extend(temp_results)
            if len(temp_results) < page_size:
                break
        results = self.__get_correct_result_count(results)
        items_collection_shell['features'] = results
        return json.dumps(items_collection_shell)



class SearchCollectionsFactory:
    CMR = 'CMR'
    UNITY = 'UNITY'
    def get_class(self, search_type):
        if search_type == SearchCollectionsFactory.CMR:
            from mdps_ds_lib.stage_in_out.search_collections_cmr import SearchCollectionsCmr
            return SearchCollectionsCmr()
        if search_type == SearchCollectionsFactory.UNITY:
            from mdps_ds_lib.stage_in_out.search_collections_unity import SearchCollectionsUnity
            return SearchCollectionsUnity()
        raise ValueError(f'unknown search_type: {search_type}')





class UploadGranulesFactory:
    UPLOAD_S3_BY_STAC_CATALOG = 'UPLOAD_S3_BY_STAC_CATALOG'
    UPLOAD_AUXILIARY_FILE_AS_GRANULE = 'UPLOAD_AUXILIARY_FILE_AS_GRANULE'

    def get_class(self, upload_type):
        if upload_type == UploadGranulesFactory.UPLOAD_S3_BY_STAC_CATALOG:
            from mdps_ds_lib.stage_in_out.upload_granules_by_complete_catalog_s3 import UploadGranulesByCompleteCatalogS3
            return UploadGranulesByCompleteCatalogS3()
        if upload_type == UploadGranulesFactory.UPLOAD_AUXILIARY_FILE_AS_GRANULE:
            from mdps_ds_lib.stage_in_out.upload_arbitrary_files_as_granules import UploadArbitraryFilesAsGranules
            return UploadArbitraryFilesAsGranules()
        raise ValueError(f'unknown search_type: {upload_type}')



class Constants:
    LOG_FORMAT = '%(asctime)s [%(levelname)s] [%(name)s::%(lineno)d] %(message)s'
    LOG_LEVEL = 'LOG_LEVEL'
    UNITY_BEARER_TOKEN = 'UNITY_BEARER_TOKEN'
    EDL_BEARER_TOKEN = 'EDL_BEARER_TOKEN'
    CLIENT_ID = 'CLIENT_ID'
    COGNITO_URL = 'COGNITO_URL'
    EDL_BASE_URL = 'EDL_BASE_URL'
    USERNAME = 'USERNAME'
    PASSWORD = 'PASSWORD'
    EDL_USERNAME = 'EDL_USERNAME'
    EDL_PASSWORD = 'EDL_PASSWORD'
    EDL_PASSWORD_TYPE = 'EDL_PASSWORD_TYPE'
    VERIFY_SSL = 'VERIFY_SSL'
    PASSWORD_TYPE = 'PASSWORD_TYPE'
    PLAIN_STR = 'PLAIN'
    BASE64_STR = 'BASE64'
    PARAM_STORE = 'PARAM_STORE'
    DAPA_API_KEY = 'DAPA_API'
    DAPA_API_PREIFX_KEY = 'DAPA_API_PREIFX_KEY'
    DAPA_API_URL_BASE = 'DAPA_API_URL_BASE'
    CORS_ORIGINS = 'CORS_ORIGINS'
    PARALLEL_COUNT = 'PARALLEL_COUNT'

    API_PREFIX = 'am-uds-dapa'
    COLLECTIONS = 'collections'
    CATALOG = 'catalog'



import logging
import time
import multiprocessing
multiprocessing.set_start_method("fork")
from multiprocessing import Process, Queue, Lock, cpu_count
from random import randint

from mdps_ds_lib.lib.processing_jobs.job_executor_abstract import JobExecutorAbstract
from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerAbstract

LOGGER = logging.getLogger(__name__)


class MultiThreadProcessorProps:
    def __init__(self, process_count: int = -1):
        self.__consumers = []
        self.__message_queue = Queue()
        self.__lock = Lock()
        self.__process_count = process_count
        if self.__process_count < 0:
            self.__process_count = cpu_count()
        self.__end_flag = 'end_of_queue'
        self.__job_manager: JobManagerAbstract = None
        self.__job_executor: JobExecutorAbstract = None

    @property
    def job_executor(self):
        return self.__job_executor

    @job_executor.setter
    def job_executor(self, val: JobExecutorAbstract):
        """
        :param val: JobExecutorAbstract
        :return: None
        """
        self.__job_executor = val
        return

    @property
    def job_manager(self):
        return self.__job_manager

    @job_manager.setter
    def job_manager(self, val: JobManagerAbstract):
        """
        :param val:
        :return: None
        """
        self.__job_manager = val
        return

    @property
    def end_flag(self):
        return self.__end_flag

    @end_flag.setter
    def end_flag(self, val):
        """
        :param val:
        :return: None
        """
        self.__end_flag = val
        return

    @property
    def consumers(self):
        return self.__consumers

    @consumers.setter
    def consumers(self, val):
        """
        :param val:
        :return: None
        """
        self.__consumers = val
        return

    @property
    def message_queue(self):
        return self.__message_queue

    @message_queue.setter
    def message_queue(self, val):
        """
        :param val:
        :return: None
        """
        self.__message_queue = val
        return

    @property
    def lock(self):
        return self.__lock

    @lock.setter
    def lock(self, val):
        """
        :param val:
        :return: None
        """
        self.__lock = val
        return

    @property
    def process_count(self):
        return self.__process_count

    @process_count.setter
    def process_count(self, val):
        """
        :param val:
        :return: None
        """
        self.__process_count = val
        return


class MultiThreadProcessor:
    def __init__(self, props=MultiThreadProcessorProps()):
        self.__props = props

    def __execute_job(self):
        while True:
            time.sleep(randint(0, 3))
            job_path = self.__props.message_queue.get()
            LOGGER.debug(f'processing: {job_path}')
            if job_path == self.__props.end_flag:
                self.__props.message_queue.put(self.__props.end_flag)
                LOGGER.debug(f'no more jobs breaking the loop')
                break
            with self.__props.lock:
                job = self.__props.job_manager.get_job_file(job_path,
                                                            lambda x: self.__props.job_executor.validate_job(x))
            if job is None:
                LOGGER.debug(f'cannot find job. continuing to next job: {job_path}')
                continue
            LOGGER.debug(f'executing job: {job}')
            result = self.__props.job_executor.execute_job(job, self.__props.lock)
            if result is True:
                LOGGER.debug(f'executed job: `{job}` successfully. removing from the processing dir')
                self.__props.job_manager.remove_from_processing(job_path)
            else:
                LOGGER.debug(f'executed job: `{job}` ends in Error. putting back to jobs dir')
                self.__props.job_manager.put_back_failed_job(job_path)
        LOGGER.debug(f'quitting __execute_job')
        return

    def start(self):
        if self.__props.job_executor is None or self.__props.job_manager is None:
            raise RuntimeError('missing job_executor or job_manager')
        LOGGER.info(f'multithread processing starting with process_count: {self.__props.process_count}')
        for i in range(self.__props.process_count):
            p = Process(target=self.__execute_job, args=())
            p.daemon = True
            self.__props.consumers.append(p)
        for c in self.__props.consumers:
            c.start()
            LOGGER.info('starting consumer pid: {}, exit_code: {}'.format(c.pid, c.exitcode))
        for k in self.__props.job_manager.get_all_job_files():
            self.__props.message_queue.put(k)
        LOGGER.debug(f'pushed all job files to queue')
        self.__props.message_queue.put(self.__props.end_flag)
        for c in self.__props.consumers:  # to check if all consumers are done processing it
            LOGGER.info('joining consumers: {}. exit_code: {}'.format(c.pid, c.exitcode))
            c.join()
            LOGGER.debug('joined')
        return



import json
import logging
import os
from abc import ABC, abstractmethod

from mdps_ds_lib.lib.utils.file_utils import FileUtils

LOGGER = logging.getLogger(__name__)


class CatalogGranulesAbstract(ABC):
    UPLOADED_FILES_JSON = 'UPLOADED_FILES_JSON'

    def __init__(self) -> None:
        super().__init__()
        self._uploaded_files_json = None

    def _retrieve_stac_json(self):
        uploaded_files_json_raw = os.environ.get(self.UPLOADED_FILES_JSON)
        try:
            self._uploaded_files_json = json.loads(uploaded_files_json_raw)
            return self
        except:
            LOGGER.debug(f'uploaded_files_json is not JSON: {uploaded_files_json_raw}. trying to see if file exists')
        if not FileUtils.file_exist(uploaded_files_json_raw):
            raise ValueError(f'missing file or not JSON: {uploaded_files_json_raw}')
        self._uploaded_files_json = FileUtils.read_json(uploaded_files_json_raw)
        if self._uploaded_files_json is None:
            raise ValueError(f'{uploaded_files_json_raw} is not JSON')
        return self

    @abstractmethod
    def catalog(self, **kwargs):
        raise NotImplementedError()




from mdps_ds_lib.lib.processing_jobs.job_manager_abstract import JobManagerProps
from mdps_ds_lib.lib.processing_jobs.job_manager_local_filesystem import JobManagerLocalFileSystem
from mdps_ds_lib.lib.processing_jobs.job_manager_memory import JobManagerMemory
from mdps_ds_lib.lib.utils.factory_abstract import FactoryAbstract


class JobManagerFactory(FactoryAbstract):
    MEMORY = 'MEMORY'
    LOCAL = 'LOCAL'
    def get_instance(self, class_type, **kwargs):
        props = JobManagerProps().load_from_json(kwargs)
        fr = class_type.upper()
        if fr == self.LOCAL:
            return JobManagerLocalFileSystem(props)
        if fr == self.MEMORY:
            return JobManagerMemory(props)
        raise ModuleNotFoundError(f'cannot find JobManagerFactory class for {class_type}')



import logging

from requests_aws4auth import AWS4Auth

from mdps_ds_lib.lib.aws.aws_cred import AwsCred
from mdps_ds_lib.lib.aws.es_middleware import ESMiddleware
from elasticsearch import Elasticsearch, RequestsHttpConnection

LOGGER = logging.getLogger(__name__)


class EsMiddlewareAws(ESMiddleware):

    def __init__(self, index, base_url, port=443) -> None:
        super().__init__(index, base_url, port)
        base_url = base_url.replace('https://', '')  # hide https
        self._index = index
        aws_cred = AwsCred()
        service = 'es'
        credentials = aws_cred.get_session().get_credentials()
        aws_auth = AWS4Auth(credentials.access_key, credentials.secret_key, aws_cred.region, service,
                            session_token=credentials.token)
        self._engine = Elasticsearch(
            hosts=[{'host': base_url, 'port': port}],
            http_auth=aws_auth,
            use_ssl=True,
            verify_certs=True,
            connection_class=RequestsHttpConnection
        )



import logging
import os
from collections import defaultdict

from pystac import Catalog, Item, Asset, Link

from mdps_ds_lib.lib.utils.file_utils import FileUtils
LOGGER = logging.getLogger(__name__)


class GranulesCatalog:
    def update_catalog(self, catalog_file_path: str, file_paths: list, rel_name: str = 'item'):
        if not FileUtils.file_exist(catalog_file_path):
            raise ValueError(f'missing file: {catalog_file_path}')
        catalog = FileUtils.read_json(catalog_file_path)
        catalog = Catalog.from_dict(catalog)
        catalog.clear_links(rel_name)
        for each_path in file_paths:
            catalog.add_link(Link('item', each_path, 'application/json'))
        return catalog.to_dict(False, False)

    def get_child_link_hrefs(self, catalog_file_path: str, rel_name: str = 'item'):
        if not FileUtils.file_exist(catalog_file_path):
            raise ValueError(f'missing file: {catalog_file_path}')
        catalog = FileUtils.read_json(catalog_file_path)
        catalog = Catalog.from_dict(catalog)
        child_links = [k.href for k in catalog.get_links(rel=rel_name)]
        catalog_dir = os.path.dirname(catalog_file_path)
        new_child_links = []
        for each_link in child_links:
            if not FileUtils.is_relative_path(each_link):
                new_child_links.append(each_link)
                continue
            new_child_links.append(os.path.join(catalog_dir, each_link))
        return new_child_links

    def get_granules_item(self, granule_stac_json) -> Item:
        if not FileUtils.file_exist(granule_stac_json):
            raise ValueError(f'missing file: {granule_stac_json}')
        granules_stac = FileUtils.read_json(granule_stac_json)
        granules_stac = Item.from_dict(granules_stac)
        return granules_stac

    def extract_assets_href(self, granules_stac: Item, dir_name: str = '') -> dict:
        try:
            self_dir = os.path.dirname(granules_stac.self_href)
        except:
            self_dir = None
        assets = defaultdict(list)
        for k, v in granules_stac.get_assets().items():
            href = v.href
            if v.roles is None or len(v.roles) < 1:
                LOGGER.warning(f'asset do not have roles: {v}')
                continue
            k = v.roles[0]
            if not FileUtils.is_relative_path(href):
                assets[k].append(href)
                continue
            if dir_name is not None and len(dir_name) > 0:
                assets[k].append(os.path.join(dir_name, href))
                continue
            if self_dir is not None and len(self_dir) > 0:
                assets[k].append(os.path.join(self_dir, href))
                continue
            assets[k].append(href)
        return assets

    def update_assets_href(self, granules_stac: Item,  new_assets: dict):
        for k, v in new_assets.items():
            if k in granules_stac.assets:
                existing_asset = granules_stac.assets.get(k)
                existing_asset.href = v
            else:
                existing_asset = Asset(v, k)
            granules_stac.add_asset(k, existing_asset)
        return self



import os
from abc import ABC, abstractmethod

from mdps_ds_lib.lib.constants import Constants


class UploadGranulesAbstract(ABC):
    RESULT_PATH_PREFIX = 'RESULT_PATH_PREFIX'  # s3 prefix
    DEFAULT_RESULT_PATH_PREFIX = 'stage_out'  # default s3 prefix
    OUTPUT_DIRECTORY = 'OUTPUT_DIRECTORY'  # To store successful & failed features json
    COLLECTION_ID_KEY = 'COLLECTION_ID'  # Need this
    STAGING_BUCKET_KEY = 'STAGING_BUCKET'  # S3 Bucket
    VERIFY_SSL_KEY = 'VERIFY_SSL'
    DELETE_FILES_KEY = 'DELETE_FILES'

    def __init__(self) -> None:
        super().__init__()
        self._collection_id = ''
        self._staging_bucket = ''
        self._result_path_prefix = ''
        self._parallel_count = int(os.environ.get(Constants.PARALLEL_COUNT, '-1'))
        self._retry_wait_time_sec = int(os.environ.get('UPLOAD_RETRY_WAIT_TIME', '30'))
        self._retry_times = int(os.environ.get('UPLOAD_RETRY_TIMES', '5'))
        self._verify_ssl = True
        self._delete_files = False

    def _set_props_from_env(self):
        missing_keys = [k for k in [self.COLLECTION_ID_KEY, self.STAGING_BUCKET_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')

        self._collection_id = os.environ.get(self.COLLECTION_ID_KEY)
        self._staging_bucket = os.environ.get(self.STAGING_BUCKET_KEY)
        self._result_path_prefix = os.environ.get(self.RESULT_PATH_PREFIX, self.DEFAULT_RESULT_PATH_PREFIX)
        self._result_path_prefix = self._result_path_prefix[:-1] if self._result_path_prefix.endswith('/') else self._result_path_prefix
        self._result_path_prefix = self._result_path_prefix[1:] if self._result_path_prefix.startswith('/') else self._result_path_prefix

        self._verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        self._delete_files = os.environ.get(self.DELETE_FILES_KEY, 'FALSE').strip().upper() == 'TRUE'
        return self

    @abstractmethod
    def upload(self, **kwargs) -> str:
        raise NotImplementedError()



class DownloadGranulesFactory:
    S3 = 'S3'
    HTTP = 'HTTP'
    DAAC = 'DAAC'

    def get_class(self, search_type):
        if search_type == DownloadGranulesFactory.S3:
            from mdps_ds_lib.stage_in_out.download_granules_s3 import DownloadGranulesS3
            return DownloadGranulesS3()
        elif search_type == DownloadGranulesFactory.DAAC:
            from mdps_ds_lib.stage_in_out.download_granules_daac import DownloadGranulesDAAC
            return DownloadGranulesDAAC()
        elif search_type == DownloadGranulesFactory.HTTP:
            from mdps_ds_lib.stage_in_out.download_granules_http import DownloadGranulesHttp
            return DownloadGranulesHttp()
        raise ValueError(f'unknown search_type: {search_type}')



from abc import ABCMeta, abstractmethod


class JobExecutorAbstract(metaclass=ABCMeta):
    @abstractmethod
    def validate_job(self, job_obj):
        return

    @abstractmethod
    def execute_job(self, job_obj, lock) -> bool:
        return



import json
import logging
import os

from mdps_ds_lib.stage_in_out.dapa_client import DapaClient
from mdps_ds_lib.stage_in_out.search_collections_abstract import SearchCollectionsAbstract

LOGGER = logging.getLogger(__name__)


class SearchCollectionsUnity(SearchCollectionsAbstract):
    COLLECTION_ID_KEY = 'COLLECTION_ID'
    DOWNLOAD_DIR_KEY = 'DOWNLOAD_DIR'

    LIMITS_KEY = 'LIMITS'
    DATE_FROM_KEY = 'DATE_FROM'
    DATE_TO_KEY = 'DATE_TO'
    VERIFY_SSL_KEY = 'VERIFY_SSL'

    def __init__(self) -> None:
        super().__init__()
        self.__collection_id = ''
        self.__date_from = ''
        self.__date_to = ''
        self.__limit = 1000
        self.__verify_ssl = True

    def __set_props_from_env(self):
        missing_keys = [k for k in [self.COLLECTION_ID_KEY] if k not in os.environ]
        if len(missing_keys) > 0:
            raise ValueError(f'missing environment keys: {missing_keys}')

        self.__collection_id = os.environ.get(self.COLLECTION_ID_KEY)
        if self.LIMITS_KEY not in os.environ:
            LOGGER.warning(f'missing {self.LIMITS_KEY}. using default: {self.__limit}')
        else:
            self.__limit = int(os.environ.get(self.LIMITS_KEY))

        self.__date_from = os.environ.get(self.DATE_FROM_KEY, '')
        self.__date_to = os.environ.get(self.DATE_TO_KEY, '')
        self.__verify_ssl = os.environ.get(self.VERIFY_SSL_KEY, 'TRUE').strip().upper() == 'TRUE'
        return self

    def search(self, **kwargs):
        self.__set_props_from_env()
        dapa_client = DapaClient().with_verify_ssl(self.__verify_ssl)
        collections_result = dapa_client.get_collection(self.__collection_id)
        return json.dumps(collections_result)



class SearchGranulesFactory:
    CMR = 'CMR'
    UNITY = 'UNITY'
    def get_class(self, search_type):
        if search_type == SearchGranulesFactory.CMR:
            from mdps_ds_lib.stage_in_out.search_granules_cmr import SearchGranulesCmr
            return SearchGranulesCmr()
        if search_type == SearchGranulesFactory.UNITY:
            from mdps_ds_lib.stage_in_out.search_granules_unity import SearchGranulesUnity
            return SearchGranulesUnity()
        raise ValueError(f'unknown search_type: {search_type}')



import fastjsonschema


class JsonValidator:
    def __init__(self, schema: dict):
        self.__schema = fastjsonschema.compile(schema)

    def validate(self, input_json: dict):
        try:
            self.__schema(input_json)
            return None
        except Exception as e:
            return str(e)



import logging

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError

from mdps_ds_lib.lib.aws.es_abstract import ESAbstract, DEFAULT_TYPE

LOGGER = logging.getLogger(__name__)


class ESMiddleware(ESAbstract):

    def __init__(self, index, base_url, port=443) -> None:
        if any([k is None for k in [index, base_url]]):
            raise ValueError(f'index or base_url is None')
        self.__index = index
        base_url = base_url.replace('https://', '')  # hide https
        # https://elasticsearch-py.readthedocs.io/en/v7.13.4/api.html#elasticsearch.Elasticsearch
        self._engine = Elasticsearch(hosts=[{'host': base_url, 'port': port, 'use_ssl': True}])

    def __validate_index(self, index):
        if index is not None:
            return index
        if self.__index is not None:
            return self.__index
        raise ValueError('index value is NULL')

    def __get_doc_dict(self, docs=None, doc_ids=None, doc_dict=None):
        if doc_dict is None and (docs is None and doc_ids is None):
            raise ValueError('must provide either doc dictionary or doc list & id list')
        if doc_dict is None:  # it comes as a list
            if len(docs) != len(doc_ids):
                raise ValueError('length of doc and id is different')
            doc_dict = {k: v for k, v in zip(doc_ids, docs)}
            pass
        return doc_dict

    def __check_errors_for_bulk(self, index_result):
        if 'errors' not in index_result or index_result['errors'] is False:
            return
        err_list = [[{'id': v['_id'], 'error': v['error']} for _, v in each.items() if 'error' in v] for each in
                    index_result['items']]
        if len(err_list) < 1:
            return
        LOGGER.exception('failed to add some items. details: {}'.format(err_list))
        return err_list

    def migrate_index_data(self, old_index, new_index, remove_old_data=True):
        if not self.has_index(old_index) or not self.has_index(new_index):
            raise ValueError(f'at least one of the indices do not exist: [{old_index}, {new_index}]')
        result = self._engine.reindex(
            body={
                "source": {
                    "index": old_index,
                    # "query": {
                    #     "match_all": {}
                    # }
                },
                "dest": {
                    "index": new_index
                }
            }
        )
        self.delete_by_query({'query': {'match_all': {}}}, old_index)
        return result

    def create_index(self, index_name, index_body):
        result = self._engine.indices.create(index=index_name, body=index_body, include_type_name=False)
        if 'acknowledged' not in result:
            return result
        return result['acknowledged']

    def get_index_mapping(self, index_name):
        try:
            result = self._engine.indices.get_mapping(index=index_name)
        except NotFoundError as e:
            return None
        return result

    def has_index(self, index_name):
        result = self._engine.indices.exists(index=index_name)
        return result

    def swap_index_for_alias(self, alias_name, old_index_name, new_index_name):
        try:
            temp_result = self._engine.indices.delete_alias(index=old_index_name, name=alias_name)
        except NotFoundError as ee:
            LOGGER.exception(f'error while unlinking {old_index_name} from {alias_name}')
            temp_result = {}
        result = self.create_alias(new_index_name, alias_name)
        return result

    def get_alias(self, alias_name):
        # /Users/wphyo/anaconda3/envs/cumulus_py_3.9/lib/python3.9/site-packages/elasticsearch-7.13.4-py3.9.egg/elasticsearch/client/indices.py
        try:
            result = self._engine.indices.get_alias(name=alias_name)
        except NotFoundError as ee:
            return {}
        return result

    def create_alias(self, index_name, alias_name):
        result = self._engine.indices.put_alias(index_name, alias_name)
        if 'acknowledged' not in result:
            return result
        return result['acknowledged']

    def delete_index(self, index_name):
        result = self._engine.indices.delete(index_name)
        if 'acknowledged' not in result:
            return result
        return result['acknowledged']

    def index_many(self, docs=None, doc_ids=None, doc_dict=None, index=None):
        # https://elasticsearch-py.readthedocs.io/en/v7.13.4/api.html#elasticsearch.Elasticsearch.bulk
        doc_dict = self.__get_doc_dict(docs, doc_ids, doc_dict)
        body = []
        for k, v in doc_dict.items():
            body.append({'index': {'_index': index, '_id': k, 'retry_on_conflict': 3}})
            body.append(v)
            pass
        index = self.__validate_index(index)
        try:
            index_result = self._engine.bulk(index=index,
                                              body=body, doc_type=DEFAULT_TYPE)
            LOGGER.info('indexed. result: {}'.format(index_result))
            return self.__check_errors_for_bulk(index_result)
        except:
            LOGGER.exception('cannot add indices with ids: {} for index: {}'.format(list(doc_dict.keys()), index))
            return doc_dict
        return

    def index_one(self, doc, doc_id, index=None):
        index = self.__validate_index(index)
        try:
            index_result = self._engine.index(index=index,
                                              body=doc, doc_type=DEFAULT_TYPE, id=doc_id)
            LOGGER.info('indexed. result: {}'.format(index_result))
            pass
        except Exception as e:
            LOGGER.exception('cannot add a new index with id: {} for index: {}'.format(doc_id, index))
            raise e
        return self

    def update_many(self, docs=None, doc_ids=None, doc_dict=None, index=None):
        doc_dict = self.__get_doc_dict(docs, doc_ids, doc_dict)
        body = []
        for k, v in doc_dict.items():
            body.append({'update': {'_index': index, '_id': k, 'retry_on_conflict': 3}})
            body.append({'doc': v, 'doc_as_upsert': True})
            pass
        index = self.__validate_index(index)
        try:
            index_result = self._engine.bulk(index=index,
                                             body=body, doc_type=DEFAULT_TYPE)
            LOGGER.info('indexed. result: {}'.format(index_result))
            return self.__check_errors_for_bulk(index_result)
        except:
            LOGGER.exception('cannot update indices with ids: {} for index: {}'.format(list(doc_dict.keys()),
                                                                                             index))
            return doc_dict
        return

    def update_one(self, doc, doc_id, index=None):
        update_body = {
            'doc': doc,
            'doc_as_upsert': True
        }
        index = self.__validate_index(index)
        try:
            update_result = self._engine.update(index=index,
                                                id=doc_id, body=update_body, doc_type=DEFAULT_TYPE)
            LOGGER.info('updated. result: {}'.format(update_result))
            pass
        except:
            LOGGER.exception('cannot update id: {} for index: {}'.format(doc_id, index))
            return None
        return self

    @staticmethod
    def get_result_size(result):
        if isinstance(result['hits']['total'], dict):  # fix for different datatype in elastic-search result
            return result['hits']['total']['value']
        else:
            return result['hits']['total']

    def query_with_scroll(self, dsl, querying_index=None):
        scroll_timeout = '30s'
        index = self.__validate_index(querying_index)
        dsl['size'] = 10000  # replacing with the maximum size to minimize number of scrolls
        params = {
            'index': index,
            'size': 10000,
            'scroll': scroll_timeout,
            'body': dsl,
        }
        first_batch = self._engine.search(**params)
        total_size = self.get_result_size(first_batch)
        current_size = len(first_batch['hits']['hits'])
        scroll_id = first_batch['_scroll_id']
        while current_size < total_size:  # need to scroll
            scrolled_result = self._engine.scroll(scroll_id=scroll_id, scroll=scroll_timeout)
            scroll_id = scrolled_result['_scroll_id']
            scrolled_result_size = len(scrolled_result['hits']['hits'])
            if scrolled_result_size == 0:
                break
            else:
                current_size += scrolled_result_size
                first_batch['hits']['hits'].extend(scrolled_result['hits']['hits'])
        return first_batch

    def query(self, dsl, querying_index=None):
        index = self.__validate_index(querying_index)
        return self._engine.search(body=dsl, index=index)

    def delete_by_query(self, dsl, querying_index=None):
        index = self.__validate_index(querying_index)
        return self._engine.delete_by_query(body=dsl, index=index)

    def __is_querying_next_page(self, targeted_size: int, current_size: int, total_size: int):
        if targeted_size < 0:
            return current_size > 0
        return current_size > 0 and total_size < targeted_size

    def query_pages(self, dsl, querying_index=None):
        if 'sort' not in dsl:
            raise ValueError('missing `sort` in DSL. Make sure sorting is unique')
        index = self.__validate_index(querying_index)
        targeted_size = dsl['size'] if 'size' in dsl else -1
        dsl['size'] = 10000  # replacing with the maximum size to minimize number of scrolls
        params = {
            'index': index,
            'size': 10000,
            'body': dsl,
        }
        LOGGER.debug(f'dsl: {dsl}')
        result_list = []
        total_size = 0
        result_batch = self._engine.search(**params)
        result_list.extend(result_batch['hits']['hits'])
        current_size = len(result_batch['hits']['hits'])
        total_size += current_size
        while self.__is_querying_next_page(targeted_size, current_size, total_size):
            params['body']['search_after'] = result_batch['hits']['hits'][-1]['sort']
            result_batch = self._engine.search(**params)
            result_list.extend(result_batch['hits']['hits'])
            current_size = len(result_batch['hits']['hits'])
            total_size += current_size
        return {
            'hits':  {
                'hits': result_list,
                'total': total_size,
            }
        }

    def query_by_id(self, doc_id, querying_index=None):
        index = self.__validate_index(querying_index)
        dsl = {
            'query': {
                'term': {'_id': doc_id}
            }
        }
        result = self._engine.search(index=index, body=dsl)
        if self.get_result_size(result) < 1:
            return None
        return result['hits']['hits'][0]['_source']



import logging
import time

from mdps_ds_lib.stage_in_out.dapa_client import DapaClient
from mdps_ds_lib.lib.cumulus_stac.item_transformer import ItemTransformer

LOGGER = logging.getLogger(__name__)


class CatalogingGranulesStatusChecker:
    def __init__(self, collection_id: str, granules_ids: list, threshold_datetime: int, delay: int, repeating_times: int, veriffy_ssl=False):
        self.__collection_id = collection_id
        self.__granules_ids = granules_ids
        self.__threshold_datetime = threshold_datetime
        self.__delay = delay
        self.__repeating_times = repeating_times if repeating_times >= 0 else 0
        self.__dapa_client = DapaClient().with_verify_ssl(veriffy_ssl)
        self.__registered_granules = {}

    def verify_one_time(self):
        granule_ids = [f"'{k}'" for k in self.__granules_ids]
        registered_granules = self.__dapa_client.get_granules(collection_id=self.__collection_id,
                                                              filters=f"id in ({','.join(granule_ids)})")
        LOGGER.debug(f'raw registered_granules: {registered_granules}')
        registered_granules = [ItemTransformer().from_stac(k) for k in registered_granules['features']]
        self.__registered_granules = {k.id: k for k in registered_granules if
                               k.datetime.timestamp() >= self.__threshold_datetime}
        LOGGER.debug(f'registered_granules after filtering: {[k for k in self.__registered_granules.keys()]}')
        LOGGER.debug(f'comparison queried v. expected: {len(self.__registered_granules)} v. {len(self.__granules_ids)}')
        missing_granules = [k for k in self.__granules_ids if k not in self.__registered_granules]
        return {
            'cataloged': len(missing_granules) < 1,
            'missing_granules': missing_granules,
            'registered_granules': [v.to_dict(include_self_link=False, transform_hrefs=False) for v in self.__registered_granules.values()]
        }

    def verify_n_times(self):
        verify_result = {
            'missing_granules': [],
            'registered_granules': []
        }
        for i in range(self.__repeating_times):
            time.sleep(self.__delay)
            verify_result = self.verify_one_time()
            LOGGER.debug(f'time {i} verification result: {verify_result}')
            if len(verify_result['missing_granules']) < 1:
                return verify_result
        return verify_result




import logging
from datetime import datetime

from pystac import Link, Collection, Extent, SpatialExtent, TemporalExtent, Summaries, Provider

from mdps_ds_lib.lib.cumulus_stac.collection_transformer import CollectionTransformer

LOGGER = logging.getLogger(__name__)


class UnityCollectionStac:
    def __init__(self):
        self.__id = ''
        self.__granule_id_extraction_regex = ''
        self.__process = ''
        self.__collection_title = ''
        self.__granule_id_regex = ''
        self.__sample_filename = ''
        self.__files = []
        self.__collection_transformer = CollectionTransformer()
        self.__provider_name = ''

    def with_provider(self, provider_name: str):
        self.__provider_name = provider_name
        return self

    def with_title(self, title: str):
        self.__collection_title = title
        return self

    def with_process(self, process: str):
        self.__process = process
        return self

    def with_id(self, collection_id: str):
        self.__id = collection_id
        if '___' not in collection_id:
            LOGGER.warning(f'no ID in {collection_id}. using 001')
            self.__id = f'{self.__id}___001'
        return self

    def with_graule_id_regex(self, granule_id_regex):
        self.__granule_id_regex = granule_id_regex
        return self

    def with_granule_id_extraction_regex(self, granule_id_extraction_regex):
        self.__granule_id_extraction_regex = granule_id_extraction_regex
        return self

    def add_file_type(self, title: str, regex: str, bucket: str, media_type: str, rel: str = 'item'):
        if rel == 'root':
            LOGGER.debug('updating media_type for rel = root')
            media_type = 'application/json'
        self.__files.append(Link(rel=rel, target=self.__collection_transformer.generate_target_link_url(regex, bucket), media_type=media_type, title=title))
        return self

    def start(self):
        # TODO validate
        stac_collection = Collection(id=self.__id,
                                     description='TODO',
                                     extent=Extent(SpatialExtent([[0, 0, 0, 0]]),
                                                   TemporalExtent([[datetime.utcnow(), datetime.utcnow()]])),
                                     title=self.__collection_title,
                                     providers=[Provider(self.__provider_name)],
                                     summaries=Summaries({
                                         'granuleId': [self.__granule_id_regex],
                                         'granuleIdExtraction': [self.__granule_id_extraction_regex],
                                         'process': [self.__process]
                                     }),
                                     )
        stac_collection.add_links(self.__files)
        new_collection = stac_collection.to_dict(include_self_link=False)
        return new_collection



import base64
import json
import logging
from copy import deepcopy

from mdps_ds_lib.lib.utils.json_validator import JsonValidator

API_GATEWAY_EVENT_SCHEMA = {
    'type': 'object',
    'required': [
        'requestContext',
        'headers',
        'httpMethod',
    ],
    'properties': {
        'requestContext': {
            'type': 'object',
            'required': [
                'path',
            ],
            'properties': {
                'path': {
                    'type': 'string'
                }
            }
        },
        'headers': {
            'type': 'object',
            'required': [
                'Host',
                'Authorization',
            ],
            'properties': {
                'Host': {
                    'type': 'string'
                },
                'Authorization': {
                    'type': 'string'
                }
            }
        },
        'httpMethod': {'type': 'string'},
    }
}
LOGGER = logging.getLogger(__name__)


SAMPLE = {
    "resource": "/am-uds-dapa/collections/{collectionId}",
    "path": "/am-uds-dapa/collections/CUMULUS_DAPA_UNIT_TEST___1665800977",
    "httpMethod": "GET",
    "headers": {
        "Accept": "*/*",
        "Accept-Encoding": "gzip, deflate",
        "Authorization": "Bearer eyJraWQiOiJzdE42WWl0eGxWZmJnY1ByRnJLWVQ1MEdjVWRIZWNBaWFKQ09peUxLVHNZPSIsImFsZyI6IlJTMjU2In0.eyJzdWIiOiIwMGRkYTdmNy1mNjE4LTRmNDMtYWFmNC1iYmM4YmExNDc2ODAiLCJjb2duaXRvOmdyb3VwcyI6WyJVbml0eV9WaWV3ZXIiLCJUZXN0X0dyb3VwIiwiVW5pdHlfQWRtaW4iXSwiaXNzIjoiaHR0cHM6XC9cL2NvZ25pdG8taWRwLnVzLXdlc3QtMi5hbWF6b25hd3MuY29tXC91cy13ZXN0LTJfRkxEeVhFMm1PIiwiY2xpZW50X2lkIjoiN2ExZmdsbTJkNTRlb2dnajEzbGNjaXZwMjUiLCJvcmlnaW5fanRpIjoiNzc3YmZkNDEtZTY1OS00M2Y0LThmYzEtNGFkOTEwNDE1NDdiIiwiZXZlbnRfaWQiOiI5YjM3OGE3ZS03OWQxLTQyMmYtODEwZi1lM2Q2Mjk2ODNiZDEiLCJ0b2tlbl91c2UiOiJhY2Nlc3MiLCJzY29wZSI6ImF3cy5jb2duaXRvLnNpZ25pbi51c2VyLmFkbWluIiwiYXV0aF90aW1lIjoxNjY1Nzc1Nzc3LCJleHAiOjE2NjU3NzkzNzcsImlhdCI6MTY2NTc3NTc3NywianRpIjoiMDk5OTA3YTMtMWVhNS00MmRiLTg4ZmMtNGU5MjQzM2FhOTA0IiwidXNlcm5hbWUiOiJ3cGh5byJ9.CiFHUKqz3q3TwW6XOPH2lkglfWO1LRk-ly-mB280GoFyGBhzSjnWHbo_U-NFmI7VsilywLMXFkif2IQ7AJSt9Cj2pja5ohrmDOFfQR_EuSSo-skSYElVaMYmIkRfVWfVa6gyByOGho5utANI0a6y9nnazvdp3ebXpdVUrNHC3yLns9-CijW-2jEvNDvEoaUZxQp06H29mcb4Iupc_SFYaVzNt3Xf_eumbyV2c0tdBEy-aRsNSOwhLXEREVJyGjHMsaSi8q2FUpHigj9ORfGtntjqESiCLCDc7wAhx6eqEZ8bfC87ck33UiFhYsamaNFHXbYTLl-uN8W5yyi8TAvJDA",
        "Host": "k3a3qmarxh.execute-api.us-west-2.amazonaws.com",
        "User-Agent": "python-requests/2.27.1",
        "X-Amzn-Trace-Id": "Root=1-6349b8e1-26c2c9e702a5cd546ba96c81",
        "X-Forwarded-For": "137.79.228.173",
        "X-Forwarded-Port": "443",
        "X-Forwarded-Proto": "https"
    },
    "multiValueHeaders": {
        "Accept": [
            "*/*"
        ],
        "Accept-Encoding": [
            "gzip, deflate"
        ],
        "Authorization": [
            "Bearer eyJraWQiOiJzdE42WWl0eGxWZmJnY1ByRnJLWVQ1MEdjVWRIZWNBaWFKQ09peUxLVHNZPSIsImFsZyI6IlJTMjU2In0.eyJzdWIiOiIwMGRkYTdmNy1mNjE4LTRmNDMtYWFmNC1iYmM4YmExNDc2ODAiLCJjb2duaXRvOmdyb3VwcyI6WyJVbml0eV9WaWV3ZXIiLCJUZXN0X0dyb3VwIiwiVW5pdHlfQWRtaW4iXSwiaXNzIjoiaHR0cHM6XC9cL2NvZ25pdG8taWRwLnVzLXdlc3QtMi5hbWF6b25hd3MuY29tXC91cy13ZXN0LTJfRkxEeVhFMm1PIiwiY2xpZW50X2lkIjoiN2ExZmdsbTJkNTRlb2dnajEzbGNjaXZwMjUiLCJvcmlnaW5fanRpIjoiNzc3YmZkNDEtZTY1OS00M2Y0LThmYzEtNGFkOTEwNDE1NDdiIiwiZXZlbnRfaWQiOiI5YjM3OGE3ZS03OWQxLTQyMmYtODEwZi1lM2Q2Mjk2ODNiZDEiLCJ0b2tlbl91c2UiOiJhY2Nlc3MiLCJzY29wZSI6ImF3cy5jb2duaXRvLnNpZ25pbi51c2VyLmFkbWluIiwiYXV0aF90aW1lIjoxNjY1Nzc1Nzc3LCJleHAiOjE2NjU3NzkzNzcsImlhdCI6MTY2NTc3NTc3NywianRpIjoiMDk5OTA3YTMtMWVhNS00MmRiLTg4ZmMtNGU5MjQzM2FhOTA0IiwidXNlcm5hbWUiOiJ3cGh5byJ9.CiFHUKqz3q3TwW6XOPH2lkglfWO1LRk-ly-mB280GoFyGBhzSjnWHbo_U-NFmI7VsilywLMXFkif2IQ7AJSt9Cj2pja5ohrmDOFfQR_EuSSo-skSYElVaMYmIkRfVWfVa6gyByOGho5utANI0a6y9nnazvdp3ebXpdVUrNHC3yLns9-CijW-2jEvNDvEoaUZxQp06H29mcb4Iupc_SFYaVzNt3Xf_eumbyV2c0tdBEy-aRsNSOwhLXEREVJyGjHMsaSi8q2FUpHigj9ORfGtntjqESiCLCDc7wAhx6eqEZ8bfC87ck33UiFhYsamaNFHXbYTLl-uN8W5yyi8TAvJDA"
        ],
        "Host": [
            "k3a3qmarxh.execute-api.us-west-2.amazonaws.com"
        ],
        "User-Agent": [
            "python-requests/2.27.1"
        ],
        "X-Amzn-Trace-Id": [
            "Root=1-6349b8e1-26c2c9e702a5cd546ba96c81"
        ],
        "X-Forwarded-For": [
            "137.79.228.173"
        ],
        "X-Forwarded-Port": [
            "443"
        ],
        "X-Forwarded-Proto": [
            "https"
        ]
    },
    "queryStringParameters": None,
    "multiValueQueryStringParameters": None,
    "pathParameters": {
        "collectionId": "CUMULUS_DAPA_UNIT_TEST___1665800977"
    },
    "stageVariables": {
        "VPCLINK": "czcxgk"
    },
    "requestContext": {
        "resourceId": "cv4985",
        "authorizer": {
            "numberKey": "123",
            "booleanKey": "true",
            "stringKey": "stringval",
            "principalId": "user",
            "integrationLatency": 303
        },
        "resourcePath": "/am-uds-dapa/collections/{collectionId}",
        "httpMethod": "GET",
        "extendedRequestId": "aAnTUHrSPHcFcGg=",
        "requestTime": "14/Oct/2022:19:30:41 +0000",
        "path": "/dev/am-uds-dapa/collections/CUMULUS_DAPA_UNIT_TEST___1665800977",
        "accountId": "884500545225",
        "protocol": "HTTP/1.1",
        "stage": "dev",
        "domainPrefix": "k3a3qmarxh",
        "requestTimeEpoch": 1665775841896,
        "requestId": "67b76cda-e49d-449e-ad3d-8ba6d42a4ce4",
        "identity": {
            "cognitoIdentityPoolId": None,
            "accountId": None,
            "cognitoIdentityId": None,
            "caller": None,
            "sourceIp": "137.79.228.173",
            "principalOrgId": None,
            "accessKey": None,
            "cognitoAuthenticationType": None,
            "cognitoAuthenticationProvider": None,
            "userArn": None,
            "userAgent": "python-requests/2.27.1",
            "user": None
        },
        "domainName": "k3a3qmarxh.execute-api.us-west-2.amazonaws.com",
        "apiId": "k3a3qmarxh"
    },
    "body": None,
    "isBase64Encoded": False
}

class LambdaApiGatewayUtils:
    def __init__(self, event: dict, default_limit: int = 10):
        self.__event = event
        self.__default_limit = default_limit
        api_gateway_event_validator_result = JsonValidator(API_GATEWAY_EVENT_SCHEMA).validate(event)
        if api_gateway_event_validator_result is not None:
            raise ValueError(f'invalid event: {api_gateway_event_validator_result}. event: {event}')

    def get_authorization_info(self):
        """
        :return:
        """
        action = self.__event['httpMethod']
        resource = self.__event['requestContext']['path']
        bearer_token = self.__event['headers']['Authorization']
        username_part = bearer_token.split('.')[1]
        jwt_decoded = base64.standard_b64decode(f'{username_part}========'.encode()).decode()
        jwt_decoded = json.loads(jwt_decoded)
        ldap_groups = jwt_decoded['cognito:groups']
        username = jwt_decoded['username']
        return {
            'username': username,
            'ldap_groups': list(set(ldap_groups)),
            'action': action,
            'resource': resource,
        }

    def __get_current_page(self):
        try:
            requesting_base_url = f"https://{self.__event['headers']['Host']}{self.__event['requestContext']['path']}"
            new_queries = deepcopy(self.__event['queryStringParameters']) if 'queryStringParameters' in self.__event and self.__event[
                'queryStringParameters'] is not None else {}
            limit = int(new_queries['limit'] if 'limit' in new_queries else self.__default_limit)
            offset = int(new_queries['offset'] if 'offset' in new_queries else 0)
            new_queries['limit'] = limit
            new_queries['offset'] = offset
            requesting_url = f"{requesting_base_url}?{'&'.join([f'{k}={v}' for k, v in new_queries.items()])}"
        except Exception as e:
            LOGGER.exception(f'error while getting current page URL')
            return f'unable to get current page URL, {str(e)}'
        return requesting_url

    def __get_next_page(self):
        try:
            requesting_base_url = f"https://{self.__event['headers']['Host']}{self.__event['requestContext']['path']}"
            new_queries = deepcopy(self.__event['queryStringParameters']) if 'queryStringParameters' in self.__event and self.__event[
                'queryStringParameters'] is not None else {}
            limit = int(new_queries['limit'] if 'limit' in new_queries else self.__default_limit)
            if limit == 0:
                return ''
            offset = int(new_queries['offset'] if 'offset' in new_queries else 0)
            offset += limit
            new_queries['limit'] = limit
            new_queries['offset'] = offset
            requesting_url = f"{requesting_base_url}?{'&'.join([f'{k}={v}' for k, v in new_queries.items()])}"
        except Exception as e:
            LOGGER.exception(f'error while getting next page URL')
            return f'unable to get next page URL, {str(e)}'
        return requesting_url

    def __get_prev_page(self):
        try:
            requesting_base_url = f"https://{self.__event['headers']['Host']}{self.__event['requestContext']['path']}"
            new_queries = deepcopy(self.__event['queryStringParameters']) if 'queryStringParameters' in self.__event and self.__event[
                'queryStringParameters'] is not None else {}
            limit = int(new_queries['limit'] if 'limit' in new_queries else self.__default_limit)
            if limit == 0:
                return ''
            offset = int(new_queries['offset'] if 'offset' in new_queries else 0)
            offset -= limit
            if offset < 0:
                offset = 0
            new_queries['limit'] = limit
            new_queries['offset'] = offset
            requesting_url = f"{requesting_base_url}?{'&'.join([f'{k}={v}' for k, v in new_queries.items()])}"
        except Exception as e:
            LOGGER.exception(f'error while getting previous page URL')
            return f'unable to get previous page URL, {str(e)}'
        return requesting_url

    def generate_pagination_links(self):
        return [
            {'rel': 'self', 'href': self.__get_current_page()},
            {'rel': 'root', 'href': f"https://{self.__event['headers']['Host']}"},
            {'rel': 'next', 'href': self.__get_next_page()},
            {'rel': 'prev', 'href': self.__get_prev_page()},
        ]



import json
import logging
import os

import requests

from mdps_ds_lib.lib.cognito_login.cognito_token_retriever import CognitoTokenRetriever
from mdps_ds_lib.lib.constants import Constants

LOGGER = logging.getLogger(__name__)


class DapaClient:
    def __init__(self, api_base_prefix=Constants.API_PREFIX):
        self.__token_retriever = CognitoTokenRetriever()
        self.__token = None
        self.__dapa_base_api = None
        self.__verify_ssl = True
        self.__api_base_prefix = api_base_prefix
        self.__get_dapa_base_api()

    def with_verify_ssl(self, verify_ssl: bool):
        self.__verify_ssl = verify_ssl
        return self

    def __get_dapa_base_api(self):
        if Constants.DAPA_API_KEY not in os.environ:
            raise ValueError(f'missing key: {Constants.DAPA_API_KEY}')
        self.__dapa_base_api = os.environ.get(Constants.DAPA_API_KEY)
        self.__dapa_base_api = self.__dapa_base_api[:-1] if self.__dapa_base_api.endswith('/') else self.__dapa_base_api
        self.__api_base_prefix = os.environ.get(Constants.DAPA_API_PREIFX_KEY) if Constants.DAPA_API_PREIFX_KEY in os.environ else self.__api_base_prefix
        return self

    def __get_token(self):
        if self.__token is None:
            self.__token = self.__token_retriever.start()
        if self.__token is None:
            raise ValueError('unable to retrieve DAPA token')
        return self

    def get_collection(self, collection_id: str):
        """
        TODO need better endpoint to get exactly 1 collection
        TODO pagination?
        :param collection_id:
        :return:
        """
        LOGGER.debug(f'getting collection details for: {collection_id}')
        self.__get_token()
        header = {'Authorization': f'Bearer {self.__token}'}
        dapa_collection_url = f'{self.__dapa_base_api}/{self.__api_base_prefix}/collections?limit=1000'
        response = requests.get(url=dapa_collection_url, headers=header, verify=self.__verify_ssl)
        if response.status_code > 400:
            raise RuntimeError(
                f'querying collections ends in error. status_code: {response.status_code}. url: {dapa_collection_url}. details: {response.text}')
        collections_result = json.loads(response.text)
        if 'features' not in collections_result:
            raise RuntimeError(f'missing features in response. invalid response: response: {collections_result}')
        collection_details = [each_collection for each_collection in collections_result['features'] if
                              collection_id == each_collection["id"]]
        if len(collection_details) < 1:
            raise RuntimeError(f'unable to find collection in DAPA')
        return collection_details[0]

    def get_all_granules(self, collection_id='*', limit=-1, date_from='', date_to=''):
        results = []
        page_size = 100 if limit < 0 or limit > 100 else limit
        offset = 0
        items_collection_shell = None
        while True:
            if 0 < limit <= len(results):
                break
            temp_results = self.get_granules(collection_id, page_size, offset, date_from, date_to)
            if items_collection_shell is None:
                items_collection_shell: dict = temp_results
            temp_results = temp_results.pop('features')
            offset += len(temp_results)
            results.extend(temp_results)
            if len(temp_results) < page_size:
                break
        if limit < 0 or limit >= len(results):
            items_collection_shell['features'] = results
            return items_collection_shell
        items_collection_shell['features'] = results[0: limit]
        return items_collection_shell

    def get_granules(self, collection_id='*', limit=1000, offset=0, date_from='', date_to='', filters=None):
        """
        TODO: pagination. getting only 1st 1k item
        :param collection_id:
        :param limit:
        :param offset:
        :param date_from:
        :param date_to:
        :return:
        """
        dapa_granules_api = f'{self.__dapa_base_api}/{self.__api_base_prefix}/collections/{collection_id}/items?limit={limit}&offset={offset}'
        if date_from != '' or date_to != '':
            dapa_granules_api = f"{dapa_granules_api}&datetime={date_from if date_from != '' else '..'}/{date_to if date_to != '' else '..'}"
        if filters is not None:
            dapa_granules_api = f'{dapa_granules_api}&filter={json.dumps(filters)}'
        LOGGER.debug(f'dapa_granules_api: {dapa_granules_api}')
        LOGGER.debug(f'getting granules for: {dapa_granules_api}')
        self.__get_token()
        header = {'Authorization': f'Bearer {self.__token}'}
        response = requests.get(url=dapa_granules_api, headers=header, verify=self.__verify_ssl)
        if response.status_code > 400:
            raise RuntimeError(
                f'querying granules ends in error. status_code: {response.status_code}. url: {dapa_granules_api}. details: {response.text}')
        granules_result = json.loads(response.text)
        if 'features' not in granules_result:
            raise RuntimeError(f'missing features in response. invalid response: response: {granules_result}')
        return granules_result

    def ingest_granules_w_cnm(self, cnm_ingest_body: dict) -> str:
        dapa_ingest_cnm_api = f'{self.__dapa_base_api}/{self.__api_base_prefix}/collections/'
        LOGGER.debug(f'getting granules for: {dapa_ingest_cnm_api}')
        self.__get_token()
        header = {
            'Authorization': f'Bearer {self.__token}',
            'Content-Type': 'application/json',
        }
        response = requests.put(url=dapa_ingest_cnm_api, headers=header, verify=self.__verify_ssl,
                                data=json.dumps(cnm_ingest_body))
        if response.status_code > 400:
            raise RuntimeError(
                f'querying granules ingestion ends in error. status_code: {response.status_code}. url: {dapa_ingest_cnm_api}. details: {response.text}')
        granules_result = json.loads(response.text)
        return granules_result



import base64
import logging
import os

from mdps_ds_lib.lib.aws.aws_param_store import AwsParamStore
from mdps_ds_lib.lib.cognito_login.cognito_login import CognitoLogin
from mdps_ds_lib.lib.constants import Constants

LOGGER = logging.getLogger()


class CognitoTokenRetriever:
    def __get_username_dwssap(self):
        username = os.environ[Constants.USERNAME]
        dwssap = os.environ[Constants.PASSWORD]
        if Constants.PASSWORD_TYPE not in os.environ:
            LOGGER.debug('no PASSWORD_TYPE set in ENV. assuming PLAIN STR')
            return username, dwssap
        dwssap_type = os.environ[Constants.PASSWORD_TYPE]
        if dwssap_type == Constants.PLAIN_STR:
            LOGGER.debug('PLAIN_STR in ENV. returning PLAIN STR')
            return username, dwssap
        if dwssap_type == Constants.BASE64_STR:
            LOGGER.debug('BASE64_STR in ENV. decoding & returning')
            username = base64.standard_b64decode(username.encode('utf-8')).decode('utf-8')
            dwssap = base64.standard_b64decode(dwssap.encode('utf-8')).decode('utf-8')
            return username, dwssap
        if dwssap_type == Constants.PARAM_STORE:
            LOGGER.debug('PARAM_STORE in ENV. retrieving value from Param Store')
            username_param_store = AwsParamStore().get_param(username)
            dwssap_param_store = AwsParamStore().get_param(dwssap)
            if username_param_store is None or dwssap_param_store is None:
                raise ValueError(f'NULL username or password from Param Store. Set the value in {username} AND {dwssap}')
            return username_param_store, dwssap_param_store
        raise ValueError(f'invalid {Constants.PASSWORD_TYPE}. value: {dwssap_type}')

    def start(self):
        if Constants.UNITY_BEARER_TOKEN in os.environ:
            LOGGER.debug('found UNITY_BEARER_TOKEN. returning UNITY_BEARER_TOKEN from ENV')
            return os.environ[Constants.UNITY_BEARER_TOKEN]
        LOGGER.debug('UNITY_BEARER_TOKEN not found. preparing to login')

        missing_mandatory_env = [k for k in [Constants.CLIENT_ID, Constants.COGNITO_URL, Constants.USERNAME, Constants.PASSWORD] if k not in os.environ]
        if len(missing_mandatory_env) > 0:
            raise ValueError(f'missing mandatory ENV for login: {missing_mandatory_env}')
        username, dwssap = self.__get_username_dwssap()
        cognito_login = CognitoLogin()\
            .with_client_id(os.environ[Constants.CLIENT_ID])\
            .with_cognito_url(os.environ[Constants.COGNITO_URL])\
            .with_verify_ssl(os.environ.get(Constants.VERIFY_SSL, 'TRUE').strip().upper() == 'TRUE')\
            .start(username, dwssap)
        if cognito_login.token is None:
            raise ValueError(f'Unity Token Retrieval failed')
        return cognito_login.token



from mdps_ds_lib.lib.aws.aws_cred import AwsCred


class AwsCognito(AwsCred):
    def __init__(self, user_pool_id: str):
        super().__init__()
        self.__cognito = self.get_client('cognito-idp')
        self.__user_pool_id = user_pool_id

    def get_groups(self, username: str):
        response = self.__cognito.admin_list_groups_for_user(
            Username=username,
            UserPoolId=self.__user_pool_id,
            Limit=60,
            # NextToken='string'
        )
        if response is None or 'Groups' not in response:
            return []
        belonged_groups = [k['GroupName'] for k in response['Groups']]
        return belonged_groups
